
from __future__ import division

import numpy as np
from scipy.fftpack import dct
from .stft import stft

def binning(S, bands):
    '''
    This function computes the sum of all columns of S in the subbands
    enumerated in bands
    '''
    B = np.zeros((S.shape[0], len(bands)), dtype=S.dtype)
    for i,b in enumerate(bands):
        B[:,i] = np.mean(S[:,b[0]:b[1]], axis=1)

    return B


def octave_bands(fc=1000, third=False):
    '''
    Create a bank of octave bands

    Parameters
    ----------
    fc : float, optional
        The center frequency
    third : bool, optional
        Use third octave bands (default False)
    '''

    div = 1
    if third == True:
        div = 3

    # Octave Bands
    fcentre = fc * ((2.0) ** (np.arange(-6*div,4*div + 1) / div))
    fd = 2**(0.5 / div);
    bands = np.array([ [f / fd, f*fd] for f in fcentre ])
    
    return bands, fcentre


def critical_bands():
    '''
    Compute the Critical bands as defined in the book:
    Psychoacoustics by Zwicker and Fastl. Table 6.1 p. 159
    '''

    # center frequencies
    fc = [50, 150, 250, 350, 450, 570, 700, 840, 1000, 1170, 1370, 1600, 1850,
          2150, 2500, 2900, 3400, 4000, 4800, 5800, 7000, 8500, 10500, 13500];
    # boundaries of the bands (e.g. the first band is from 0Hz to 100Hz 
    # with center 50Hz, fb[0] to fb[1], center fc[0]
    fb = [0,  100, 200, 300, 400, 510, 630, 770, 920, 1080, 1270, 1480, 1720,
          2000, 2320, 2700, 3150, 3700, 4400, 5300, 6400, 7700, 9500, 12000, 15500];

    # now just make pairs
    bands = [ [fb[j], fb[j+1]] for j in range(len(fb)-1) ]

    return np.array(bands), fc


def bands_hz2s(bands_hz, Fs, N, transform='dft'):
    '''
    Converts bands given in Hertz to samples with respect to a given sampling
    frequency Fs and a transform size N an optional transform type is used to
    handle DCT case.
    '''

    # set the bin width
    if (transform == 'dct'):
        B = Fs/2/N
    else:
        B = Fs/N

    # upper limit of the frequency range
    limit = min(Fs/2, bands_hz[-1,1])

    # Convert from Hertz to samples for all bands
    bands_s = [ np.around(band/B)  for band in bands_hz if band[0] <= limit]

    # Last band ends at N/2
    bands_s[-1][1] = N/2

    # remove duplicate, if any, (typically, if N is small and Fs is large)
    j = 0
    while (j < len(bands_s)-1):
        if (bands_s[j][0] == bands_s[j+1][0]):
            bands_s.pop(j)
        else:
            j += 1

    return np.array(bands_s, dtype=np.int)

def melscale(f):
    ''' Converts f (in Hertz) to the melscale defined according to Huang-Acero-Hon (2.6) '''
    return 1125.*np.log(1+f/700.)

def invmelscale(b):
    ''' Converts from melscale to frequency in Hertz according to Huang-Acero-Hon (6.143) '''
    return 700.*(np.exp(b/1125.)-1)

def melfilterbank(M, N, fs=1, fl=0., fh=0.5):
    '''
    Returns a filter bank of triangular filters spaced according to mel scale

    We follow Huang-Acera-Hon 6.5.2

    Parameters
    ----------
    M : (int)
        The number of filters in the bank
    N : (int)
        The length of the DFT
    fs : (float) optional
        The sampling frequency (default 8000)
    fl : (float)
        Lowest frequency in filter bank as a fraction of fs (default 0.)
    fh : (float)
        Highest frequency in filter bank as a fraction of fs (default 0.5)

    Returns
    -------
    An M times int(N/2)+1 ndarray that contains one filter per row
    '''

    # all center frequencies of the filters
    f = (N/fs)*invmelscale( melscale(fl*fs) + (
            np.arange(M+2)*(melscale(fh*fs)-melscale(fl*fs))/(M+1) )
        )

    # Construct the triangular filter bank
    H = np.zeros((M, N//2+1))
    k = np.arange(N//2+1)
    for m in range(1,M+1):
        I = np.where(np.logical_and(f[m-1] < k, k < f[m]))
        H[m-1,I] = 2 * (k[I]-f[m-1]) / ((f[m+1]-f[m-1]) * (f[m]-f[m-1]))
        I = np.where(np.logical_and(f[m] <= k, k < f[m+1]))
        H[m-1,I] = 2 * (f[m+1]-k[I]) / ((f[m+1]-f[m-1]) * (f[m+1]-f[m]))

    return H


def mfcc(x, L=128, hop=64, M=14, fs=8000, fl=0., fh=0.5):
    '''
    Computes the Mel-Frequency Cepstrum Coefficients (MFCC) according
    to the description by Huang-Acera-Hon 6.5.2 (2001)
    The MFCC are features mimicing the human perception usually
    used for some learning task.

    This function will first split the signal into frames, overlapping
    or not, and then compute the MFCC for each frame.

    Parameters
    ----------
    x : (nd-array)
        Input signal
    L : (int)
        Frame size (default 128)
    hop : (int)
        Number of samples to skip between two frames (default 64)
    M : (int)
        Number of mel-frequency filters (default 14)
    fs : (int)
        Sampling frequency (default 8000)
    fl : (float)
        Lowest frequency in filter bank as a fraction of fs (default 0.)
    fh : (float)
        Highest frequency in filter bank as a fraction of fs (default 0.5)

    Return
    ------
    The MFCC of the input signal
    '''

    # perform STFT, X contains frames in rows
    X = stft(x, L, hop, transform=np.fft.rfft)

    # get and apply the mel filter bank
    # and compute log energy
    H = melfilterbank(M, L, fs=fs, fl=fl, fh=fh)
    S = np.log(np.dot(H, np.abs(X.T)**2))

    # Now take DCT of the result
    C = dct(S, type=2, n=M, axis=0)

    return C





import numpy as np
from scipy.signal import fftconvolve, correlate

def trinicon(signals):
    '''
    Implementation of the TRINICON Blind Source Separation algorithm as described in

    Aichner, R., Buchner, H., Yan, F., & Kellermann, W. (2006). 
    A real-time blind source separation scheme and its application to reverberant and noisy acoustic environments. 
    Signal Processing, 86(6), 1260-1277. doi:10.1016/j.sigpro.2005.06.022

    Specifically, adaptation of the pseudo-code from Table 1.

    The implementation is hard-coded for 2 output channels.
    '''

    P = signals.shape[0] # number of microphones
    Q = 2                # number of output channels

    K = 8                # Number of successive blocks processed at the same time
    L = 4096             # Filter length
    N = 2*L              # Block length
    alpha_on = K         # online overlap factor
    alpha_off = 1        # offline overlap factor (not used here)

    j_max = 10           # number of offline iterations

    delta_max = 1e-4     # regularization parameter, this sets the maximum value of the regularization term
    sigma2_0  = 1e-7     # regularization parameter, this sets the reference (machine?) noise level in the regularization

    mu = 0.0010          # offline update step size
    lambd_a = 0.2        # online forgetting factor

    # the filters
    w = np.zeros((P,Q,L))
    w[:P/2,0,L/2] = 1
    w[P/2:,1,L/2] = 1

    hop = K*L/alpha_on

    # pad with zeros to have a whole number of online blocks
    if signals.shape[1] % hop != 0:
        signals = np.concatenate((signals, np.zeros((P, hop - (signals.shape[0]%hop)))), axis=1)

    S = signals.shape[1] # total signal length
    M = S / hop          # number of online blocks

    y = np.zeros((Q,S))    # the processed output signal

    m = 1               # online block index
    while m <= M:        # online loop
        
        # new chunk of input signal
        x = np.zeros((P,K*L+N))
        if m*hop > S:
            # we need some zero padding at the back
            le = S - (m-1)*hop + N
            x[:le] = signals[:,m*hop-K*L-N:]
        if m*hop >= K*L+N:
            x = signals[:,m*hop-K*L-N:m*hop]
        else:
            # we need some zero padding at the beginning
            x[:,-m*hop:] = signals[:,:m*hop]

        # use filter from previous iteration to initialize offline part
        w_new = w.copy()

        for j in range(j_max):     # offline update loop

            y_c = np.zeros((Q,K*L+N-L))  # c stands for chunk
            y_blocks = np.zeros((Q,K,N))

            for q in range(Q):
                # convolve with filters
                for p in range(P):
                    # We discard the 'oldest' output of the convolution according
                    # to the filter matrix definition (6) in the paper
                    y_c[q,:] += fftconvolve(x[p,:], w_new[p,q,:], mode='valid')[1:]

                # split into smaller blocks
                for i in range(K):
                    y_blocks[q,i,:] = y_c[q,i*L:i*L+N]

            # blocks energy
            sigma2 = np.sum(y_blocks**2, axis=2)

            # cross-correlations
            r_cross = np.zeros((Q,K,2*L-1))
            for i in range(K):
                y0 = y_c[0,i*L:i*L+N]
                y1 = y_c[1,i*L:i*L+N]
                r = fftconvolve(y1, y0[::-1], mode='full')
                r_cross[0,i,:] = r[N-L:N+L-1]         # r_y1y0
                r_cross[1,i,:] = r_cross[0,i,::-1]    # r_y0y1 by symmetry is just r_y1y0 reversed

            # regularization term
            delta = delta_max*np.exp(-sigma2/sigma2_0)

            # offline update
            delta_w = np.zeros((P,Q,L))
            for q in range(Q):
                for p in range(P):
                    for i in range(K):
                        # this implements the row-wise sylvester constraint as explained in Fig. 4 (b) of paper
                        delta_w[p,q,:] += fftconvolve(r_cross[q,i,:]/(sigma2[q,i]+delta[q,i]), w_new[p,1-q,::-1], mode='valid')[::-1]
                    delta_w[p,q,:] /= K

            w_new = w_new - mu*delta_w

        # online update
        w = lambd_a*w + (1-lambd_a)*w_new

        # compute output signal
        for q in range(Q):
            for p in range(P):
                y[q,(m-1)*hop:m*hop] += fftconvolve(x[p,-hop-L+1:], w[p,q,:], mode='valid')
        
        # next block
        m += 1

    return y




import ctypes as _ctypes
import os
import glob

path = os.path.dirname(__file__)

try:
    # we need the matching because python3 appends some os info to the name
    match_files = glob.glob(path + "/libroom*so")
    libroom = _ctypes.cdll.LoadLibrary(match_files[0])
    libroom_available = True
except:
    libroom = False
    libroom_available = False

from .libroom_wrapper import *



# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

'''A collection of windowing functions.'''

import numpy as np

# cosine window function
def cosine(N, flag='asymmetric', length='full'):

    # first choose the indexes of points to compute
    if (length == 'left'):     # left side of window
        t = np.arange(0, N / 2)
    elif(length == 'right'):   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if (flag == 'symmetric' or flag == 'mdct'):
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = np.cos(np.pi * (t - 0.5)) ** 2

    # make the window respect MDCT condition
    if (flag == 'mdct'):
        w **= 2
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# triangular window function
def triang(N, flag='asymmetric', length='full'):

    # first choose the indexes of points to compute
    if (length == 'left'):     # left side of window
        t = np.arange(0, N / 2)
    elif(length == 'right'):   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if (flag == 'symmetric' or flag == 'mdct'):
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = 1. - np.abs(2. * t - 1.)

    # make the window respect MDCT condition
    if (flag == 'mdct'):
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# hann window function
def hann(N, flag='asymmetric', length='full'):

    # first choose the indexes of points to compute
    if (length == 'left'):     # left side of window
        t = np.arange(0, N / 2)
    elif(length == 'right'):   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if (flag == 'symmetric' or flag == 'mdct'):
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = 0.5 * (1 - np.cos(2 * np.pi * t))

    # make the window respect MDCT condition
    if (flag == 'mdct'):
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# Blackman-Harris window
def blackman_harris(N, flag='asymmetric', length='full'):

    # coefficients
    a = np.array([.35875, .48829, .14128, .01168])

    # first choose the indexes of points to compute
    if (length == 'left'):     # left side of window
        t = np.arange(0, N / 2)
    elif(length == 'right'):   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if (flag == 'symmetric'):
        t = t / float(N - 1)
    else:
        t = t / float(N)

    pi = np.pi
    w = a[0] - a[1]*np.cos(2*pi*t) + a[2]*np.cos(4*pi*t) + a[3]*np.cos(6*pi*t)

    return w

# Rectangular window function
def rect(N):
    return np.ones(N)



# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

import numpy as np
from scipy import signal

from .parameters import constants, eps


def to_16b(signal):
    '''
    converts float 32 bit signal (-1 to 1) to a signed 16 bits representation
    No clipping in performed, you are responsible to ensure signal is within
    the correct interval.
    '''
    return ((2**15-1)*signal).astype(np.int16)

def clip(signal, high, low):
    '''Clip a signal from above at high and from below at low.'''
    s = signal.copy()

    s[np.where(s > high)] = high
    s[np.where(s < low)] = low

    return s


def normalize(signal, bits=None):
    '''
    normalize to be in a given range. The default is to normalize the maximum
    amplitude to be one. An optional argument allows to normalize the signal
    to be within the range of a given signed integer representation of bits.
    '''

    s = signal.copy()

    s /= np.abs(s).max()

    # if one wants to scale for bits allocated
    if bits is not None:
        s *= 2 ** (bits - 1)
        s = clip(signal, 2 ** (bits - 1) - 1, -2 ** (bits - 1))

    return s


def normalize_pwr(sig1, sig2):
    '''Normalize sig1 to have the same power as sig2.'''

    # average power per sample
    p1 = np.mean(sig1 ** 2)
    p2 = np.mean(sig2 ** 2)

    # normalize
    return sig1.copy() * np.sqrt(p2 / p1)


def highpass(signal, Fs, fc=None, plot=False):
    ''' Filter out the really low frequencies, default is below 50Hz '''

    if fc is None:
        fc = constants.get('fc_hp')

    # have some predefined parameters
    rp = 5  # minimum ripple in dB in pass-band
    rs = 60   # minimum attenuation in dB in stop-band
    n = 4    # order of the filter
    type = 'butter'

    # normalized cut-off frequency
    wc = 2. * fc / Fs

    # design the filter
    from scipy.signal import iirfilter, lfilter, freqz
    b, a = iirfilter(n, Wn=wc, rp=rp, rs=rs, btype='highpass', ftype=type)

    # plot frequency response of filter if requested
    if (plot):
        import matplotlib.pyplot as plt
        w, h = freqz(b, a)

        plt.figure()
        plt.title('Digital filter frequency response')
        plt.plot(w, 20 * np.log10(np.abs(h)))
        plt.title('Digital filter frequency response')
        plt.ylabel('Amplitude Response [dB]')
        plt.xlabel('Frequency (rad/sample)')
        plt.grid()

    # apply the filter
    signal = lfilter(b, a, signal.copy())

    return signal


def time_dB(signal, Fs, bits=16):
    '''
    Compute the signed dB amplitude of the oscillating signal
    normalized wrt the number of bits used for the signal.
    '''

    import matplotlib.pyplot as plt

    # min dB (least significant bit in dB)
    lsb = -20 * np.log10(2.) * (bits - 1)

    # magnitude in dB (clipped)
    pos = clip(signal, 2. ** (bits - 1) - 1, 1.) / 2. ** (bits - 1)
    neg = -clip(signal, -1., -2. ** (bits - 1)) / 2. ** (bits - 1)

    mag_pos = np.zeros(signal.shape)
    Ip = np.where(pos > 0)
    mag_pos[Ip] = 20 * np.log10(pos[Ip]) + lsb + 1

    mag_neg = np.zeros(signal.shape)
    In = np.where(neg > 0)
    mag_neg[In] = 20 * np.log10(neg[In]) + lsb + 1

    plt.plot(np.arange(len(signal)) / float(Fs), mag_pos - mag_neg)
    plt.xlabel('Time [s]')
    plt.ylabel('Amplitude [dB]')
    plt.axis('tight')
    plt.ylim(lsb-1, -lsb+1)

    # draw ticks corresponding to decibels
    div = 20
    n = int(-lsb/div)+1
    yticks = np.zeros(2*n)
    yticks[:n] = lsb - 1 + np.arange(0, n*div, div)
    yticks[n:] = -lsb + 1 - np.arange((n-1)*div, -1, -div)
    yticklabels = np.zeros(2*n)
    yticklabels = range(0, -n*div, -div) + range(-(n-1)*div, 1, div)
    plt.setp(plt.gca(), 'yticks', yticks)
    plt.setp(plt.gca(), 'yticklabels', yticklabels)

    plt.setp(plt.getp(plt.gca(), 'ygridlines'), 'ls', '--')


def spectrum(signal, Fs, N):

    from .stft import stft, spectroplot
    from .windows import hann

    F = stft(signal, N, N / 2, win=hann(N))
    spectroplot(F.T, N, N / 2, Fs)


def dB(signal, power=False):
    if power is True:
        return 10*np.log10(np.abs(signal))
    else:
        return 20*np.log10(np.abs(signal))


def compare_plot(signal1, signal2, Fs, fft_size=512, norm=False, equal=False, title1=None, title2=None):

    import matplotlib.pyplot as plt

    td_amp = np.maximum(np.abs(signal1).max(), np.abs(signal2).max())

    if norm:
        if equal:
            signal1 /= np.abs(signal1).max()
            signal2 /= np.abs(signal2).max()
        else:
            signal1 /= td_amp
            signal2 /= td_amp
        td_amp = 1.

    plt.subplot(2, 2, 1)
    plt.plot(np.arange(len(signal1))/float(Fs), signal1)
    plt.axis('tight')
    plt.ylim(-td_amp, td_amp)
    if title1 is not None:
        plt.title(title1)

    plt.subplot(2, 2, 2)
    plt.plot(np.arange(len(signal2))/float(Fs), signal2)
    plt.axis('tight')
    plt.ylim(-td_amp, td_amp)
    if title2 is not None:
        plt.title(title2)

    from .stft import stft, spectroplot
    from .windows import hann

    F1 = stft.stft(signal1, fft_size, fft_size / 2, win=windows.hann(fft_size))
    F2 = stft.stft(signal2, fft_size, fft_size / 2, win=windows.hann(fft_size))

    # try a fancy way to set the scale to avoid having the spectrum
    # dominated by a few outliers
    p_min = 1
    p_max = 99.5
    all_vals = np.concatenate((dB(F1+eps), dB(F2+eps))).flatten()
    vmin, vmax = np.percentile(all_vals, [p_min, p_max])

    cmap = 'jet'
    interpolation = 'sinc'

    plt.subplot(2, 2, 3)
    stft.spectroplot(F1.T, fft_size, fft_size / 2, Fs, vmin=vmin, vmax=vmax,
                    cmap=plt.get_cmap(cmap), interpolation=interpolation)

    plt.subplot(2, 2, 4)
    stft.spectroplot(F2.T, fft_size, fft_size / 2, Fs, vmin=vmin, vmax=vmax,
            cmap=plt.get_cmap(cmap), interpolation=interpolation)


def real_spectrum(signal, axis=-1, **kwargs):

    import matplotlib.pyplot as plt

    S = np.fft.rfft(signal, axis=axis)
    f = np.arange(S.shape[axis])/float(2*S.shape[axis])

    plt.subplot(2,1,1)
    P = dB(S)
    plt.plot(f, P, **kwargs)

    plt.subplot(2,1,2)
    phi = np.unwrap(np.angle(S))
    plt.plot(f, phi, **kwargs)



def convmtx(x, n):
    '''
    Create a convolution matrix H for the vector x of size len(x) times n.
    Then, the result of np.dot(H,v) where v is a vector of length n is the same
    as np.convolve(x, v).
    '''

    import scipy as s

    c = np.concatenate((x, np.zeros(n-1)))
    r = np.zeros(n)

    return s.linalg.toeplitz(c, r)


def prony(x, p, q):
    '''
    Prony's Method from Monson H. Hayes' Statistical Signal Processing, p. 154

    Parameters
    ----------

    x: 
        signal to model
    p: 
        order of denominator
    q: 
        order of numerator

    Returns
    -------

    a: 
        numerator coefficients
    b: 
        denominator coefficients
    err: the squared error of approximation
    '''

    nx = x.shape[0]

    if (p+q >= nx):
        raise NameError('Model order too large')

    X = convmtx(x, p+1)

    Xq = X[q:nx+p-1, 0:p]

    a = np.concatenate((np.ones(1), -np.linalg.lstsq(Xq, X[q+1:nx+p, 0])[0]))
    b = np.dot(X[0:q+1, 0:p+1], a)

    err = np.inner(np.conj(x[q+1:nx]), np.dot(X[q+1:nx, :p+1], a))

    return a, b, err


def shanks(x, p, q):
    '''
    Shank's Method from Monson H. Hayes' Statistical Signal Processing, p. 154

    Parameters
    ----------
    x: 
        signal to model
    p: 
        order of denominator
    q: 
        order of numerator

    Returns
    -------
    a: 
        numerator coefficients
    b: 
        denominator coefficients
    err: 
        the squared error of approximation
    '''

    from scipy import signal

    nx = x.shape[0]

    if (p+q >= nx):
        raise NameError('Model order too large')

    a = prony(x, p, q)[0]

    u = np.zeros(nx)
    u[0] = 1.

    g = signal.lfilter(np.ones(1), a, u)

    G = convmtx(g, q+1)
    b = np.linalg.lstsq(G[:nx, :], x)[0]
    err = np.inner(np.conj(x), x) - np.inner(np.conj(x), np.dot(G[:nx, :q+1], b))

    return a, b, err


def low_pass_dirac(t0, alpha, Fs, N):
    '''
    Creates a vector containing a lowpass Dirac of duration T sampled at Fs
    with delay t0 and attenuation alpha.

    If t0 and alpha are 2D column vectors of the same size, then the function
    returns a matrix with each line corresponding to pair of t0/alpha values.
    '''

    return alpha*np.sinc(np.arange(N) - Fs*t0)


def fractional_delay(t0):
    '''
    Creates a fractional delay filter using a windowed sinc function.
    The length of the filter is fixed by the module wide constant
    `frac_delay_length` (default 81).

    Parameters
    ----------
    t0: float
        The delay in fraction of sample. Typically between 0 and 1.

    Returns
    -------
    A fractional delay filter with specified delay.
    '''

    N = constants.get('frac_delay_length')

    return np.hanning(N)*np.sinc(np.arange(N) - (N-1)/2 - t0)


def fractional_delay_filter_bank(delays):
    '''
    Creates a fractional delay filter bank of windowed sinc filters

    Parameters
    ----------
    delays: 1d narray
        The delays corresponding to each filter in fractional samples
        
    Returns
    -------
    An ndarray where the ith row contains the fractional delay filter
    corresponding to the ith delay. The number of columns of the matrix
    is proportional to the maximum delay.
    '''

    # constants and lengths
    N = delays.shape[0]
    L = constants.get('frac_delay_length')
    filter_length = L + int(np.ceil(delays).max())

    # subtract the minimum delay, so that all delays are positive
    delays -= delays.min()

    # allocate a flat array for the filter bank that we'll reshape at the end
    bank_flat = np.zeros(N * filter_length)

    # separate delays in integer and fractional parts
    di = np.floor(delays).astype(np.int)
    df = delays - di

    # broadcasting tricks to compute at once all the locations
    # and sinc times that must be computed
    T = np.tile(np.arange(L), (N, 1))
    indices = (T + (di[:,None] + filter_length * np.arange(N)[:,None]))
    sinc_times = (T - df[:,None] - (L - 1) / 2)

    # we'll need to window also all the sincs at once
    windows = np.tile(np.hanning(L), N)

    # compute all sinc with one call
    bank_flat[indices.ravel()] = windows * np.sinc(sinc_times.ravel())

    return np.reshape(bank_flat, (N, -1))



def levinson(r, b):
    '''
    levinson(r,b)

    Solve a system of the form Rx=b where R is hermitian toeplitz matrix and b
    is any vector using the generalized Levinson recursion as described in M.H.
    Hayes, Statistical Signal Processing and Modelling, p. 268.

    Parameters
    ----------
    r: 
        First column of R, toeplitz hermitian matrix.
    b: 
        The right-hand argument. If b is a matrix, the system is solved
        for every column vector in b.

    Returns
    -------
    The solution of the linear system Rx = b.
    '''

    p = b.shape[0]

    a = np.array([1])
    x = b[np.newaxis, 0, ]/r[0]
    epsilon = r[0]

    for j in np.arange(1, p):

        g = np.sum(np.conj(r[1:j+1])*a[::-1])
        gamma = -g/epsilon
        a = np.concatenate((a, np.zeros(1))) + gamma*np.concatenate((np.zeros(1), np.conj(a[::-1])))
        epsilon = epsilon*(1 - np.abs(gamma)**2)
        delta = np.dot(np.conj(r[1:j+1]), np.flipud(x))
        q = (b[j, ] - delta)/epsilon
        x = np.concatenate((x, np.zeros(1 if len(b.shape) == 1 else (1, b.shape[1]))), axis=0) + q*np.conj(a[::-1, np.newaxis])

    return x

def goertzel(x, k):
    ''' Goertzel algorithm to compute DFT coefficients '''

    N = x.shape[0]
    f = k/float(N)

    a = np.r_[1., -2.*np.cos(2.*np.pi*f), 1.]
    b = np.r_[1]
    s = signal.lfilter(b,a, x)
    y = np.exp(2j*np.pi*f)*s[-1] - s[-2]

    return y
    
    
'''
GEOMETRY UTILITIES
'''
def angle_from_points(x1, x2):
    return np.angle((x1[0, 0]-x2[0, 0]) + 1j*(x1[1, 0] - x2[1, 0]))





import numpy as np

def phat(x1, x2):

    N1 = x1.shape[0]
    N2 = x2.shape[0]

    N = N1 + N2 - 1

    X1 = np.fft.rfft(x1, n=N)
    X1 /= np.abs(X1)

    X2 = np.fft.rfft(x2, n=N)
    X2 /= np.abs(X2)

    r_12 = np.fft.irfft(X1*np.conj(X2), n=N)

    '''
    import matplotlib.pyplot as plt 
    plt.figure()
    plt.plot(r_12)
    plt.show()
    '''

    i = np.argmax(np.abs(r_12))

    if i < N1:
        return i
    else:
        return i - N1 - N2 + 1

def correlation(x1, x2):

    N1 = x1.shape[0]
    N2 = x2.shape[0]

    N = N1 + N2 - 1

    x1_p = np.zeros(N)
    x1_p[:N1] = x1
    x2_p = np.zeros(N)
    x2_p[:N2] = x2

    X1 = np.fft.fft(x1_p)

    X2 = np.fft.fft(x2_p)

    r_12 = np.real(np.fft.ifft(X1*np.conj(X2)))

    '''
    import matplotlib.pyplot as plt 
    plt.figure()
    plt.plot(np.real(r_12))
    plt.plot(np.imag(r_12))
    plt.show()
    '''

    i = np.argmax(r_12)

    if i < N1:
        return i
    else:
        return i - N1 - N2 + 1


def delay_estimation(x1, x2, L):
    '''
    Estimate the delay between x1 and x2.
    L is the block length used for phat
    '''

    K = np.minimum(x1.shape[0], x2.shape[0])/L

    delays = np.zeros(K)
    for k in range(K):
        delays[k] = phat(x1[k*L:(k+1)*L], x2[k*L:(k+1)*L])

    return int(np.median(delays))


def time_align(ref, deg, L=4096):
    '''
    return a copy of deg time-aligned and of same-length as ref.
    L is the block length used for correlations.
    '''

    # estimate delay of signal
    from numpy import zeros, minimum
    delay = delay_estimation(ref, deg, L)

    # time-align with reference segment for error metric computation
    sig = zeros(ref.shape[0])
    if (delay >= 0):
        length = minimum(deg.shape[0], ref.shape[0]-delay)
        sig[delay:length+delay] = deg[:length]
    else:
        length = minimum(deg.shape[0]+delay, ref.shape[0])
        sig = zeros(ref.shape)
        sig[:length] = deg[-delay:-delay+length]

    return sig




# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

'''Collection of spectral estimation methods.'''
from __future__ import division

import sys
import numpy as np
from scipy.signal import correlate as correlate
import matplotlib.pyplot as plt

from numpy.lib.stride_tricks import as_strided

# a routine for long convolutions using overlap add method


def overlap_add(in1, in2, L):

    # set the shortest sequence as the filter
    if (len(in1) > len(in2)):
        x = in1
        h = in2
    else:
        h = in1
        x = in2

    # filter length
    M = len(h)

    # FFT size
    N = L + M - 1

    # frequency domain filter (zero-padded)
    H = np.fft.rfft(h, N)

    # prepare output signal
    ylen = int(np.ceil(len(x) / float(L)) * L + M - 1)
    y = np.zeros(ylen)

    # overlap add
    i = 0
    while (i < len(x)):
        y[i:i + N] += np.fft.irfft(np.fft.rfft(x[i:i + L], N) * H, N)
        i += L

    return y[:len(x) + M - 1]


# Nicely plot the spectrogram
def spectroplot(Z, N, hop, fs, fdiv=None, tdiv=None,
                vmin=None, vmax=None, cmap=None, interpolation='none', colorbar=True):

    plt.imshow(
        20 * np.log10(np.abs(Z[:N // 2 + 1, :])),
        aspect='auto',
        origin='lower',
        vmin=vmin, vmax=vmax, cmap=cmap, interpolation=interpolation)

    # label y axis correctly
    plt.ylabel('Freq [Hz]')
    yticks = plt.getp(plt.gca(), 'yticks')
    plt.setp(plt.gca(), 'yticklabels', np.round(yticks / float(N) * fs))
    if (fdiv is not None):
        tick_lbls = np.arange(0, fs / 2, fdiv)
        tick_locs = tick_lbls * N / fs
        plt.yticks(tick_locs, tick_lbls)

    # label x axis correctly
    plt.xlabel('Time [s]')
    xticks = plt.getp(plt.gca(), 'xticks')
    plt.setp(plt.gca(), 'xticklabels', xticks / float(fs) * hop)
    if (tdiv is not None):
        unit = float(hop) / fs
        length = unit * Z.shape[1]
        tick_lbls = np.arange(0, int(length), tdiv)
        tick_locs = tick_lbls * fs / hop
        plt.xticks(tick_locs, tick_lbls)

    if colorbar is True:
        plt.colorbar(orientation='horizontal')

# A more general implementation of STFT


def stft(x, L, hop, transform=np.fft.fft, win=None, zp_back=0, zp_front=0):
    '''
    Parameters:
    -----------
    x: 
        input signal
    L: 
        frame size
    hop: 
        shift size between frames
    transform: 
        the transform routine to apply (default FFT)
    win: 
        the window to apply (default None)
    zp_back: 
        zero padding to apply at the end of the frame
    zp_front: 
        zero padding to apply at the beginning of the frame

    Returns:
    --------
    The STFT of x
    '''

    # the transform size
    N = L + zp_back + zp_front

    # window needs to be same size as transform
    if (win is not None and len(win) != N):
        print('Window length need to be equal to frame length + zero padding.')
        sys.exit(-1)

    # reshape
    new_strides = (hop * x.strides[0], x.strides[0])
    new_shape = ((len(x) - L) // hop + 1, L)
    y = as_strided(x, shape=new_shape, strides=new_strides)

    # add the zero-padding
    y = np.concatenate(
        (np.zeros(
            (y.shape[0], zp_front)), y, np.zeros(
            (y.shape[0], zp_back))), axis=1)

    # apply window if needed
    if (win is not None):
        y = win * y
        # y = np.expand_dims(win, 0)*y

    # transform along rows
    Z = transform(y, axis=1)

    # apply transform
    return Z


# inverse STFT
def istft(X, L, hop, transform=np.fft.ifft, win=None, zp_back=0, zp_front=0):

    # the transform size
    N = L + zp_back + zp_front

    # window needs to be same size as transform
    if (win is not None and len(win) != N):
        print('Window length need to be equal to frame length + zero padding.')
        sys.exit(-1)

    # inverse transform
    iX = transform(X, axis=1)
    if (iX.dtype == 'complex128'):
        iX = np.real(iX)

    # apply synthesis window if necessary
    if (win is not None):
        iX *= win

    # create output signal
    x = np.zeros(X.shape[0] * hop + (L - hop) + zp_back + zp_front)

    # overlap add
    for i in range(X.shape[0]):
        x[i * hop:i * hop + N] += iX[i]

    return x


# FreqVec: given FFT size and sampling rate, returns a vector of real
# frequencies
def freqvec(N, fs, centered=False):
    '''
    N: FFT length
    fs: sampling rate of the signal
    shift: False if the DC is at the beginning, True if the DC is centered
    '''

    # Create a centered vector. The (1-N%2) is to correct for even/odd length
    vec = np.arange(-N // 2 + (1 - N % 2), N // 2 + 1) * float(fs) / float(N)

    # Shift positive/negative frequencies if needed. Again (1-N%2) for
    # even/odd length
    if centered:
        return vec
    else:
        return np.concatenate((vec[N // 2 - (1 - N % 2):], vec[0:N // 2 - 1]))



# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

import numpy as np

from . import geometry as geom



class Wall(object):
    '''
    This class represents a wall instance. A room instance is formed by these.
    
    :attribute corners: (np.array dim 2x2 or 3xN, N>2) endpoints forming the wall
    :attribute absorption: (float) attenuation reflection factor
    :attribute name: (string) name given to the wall, which can be reused to reference it in the Room object
    :attribute normal: (np.array dim 2 or 3) normal vector pointing outward the room
    :attribute dim: (int) dimension of the wall (2 or 3, meaning 2D or 3D)
    '''
    
    
    def __init__(
        self,
        corners,
        absorption = 1.,
        name = None):
        
        self.corners = np.array(corners, order='F', dtype=np.float32)
        self.absorption = absorption

        # set first corner as origin of plane
        self.plane_point = np.array(self.corners[:,0])

        if (self.corners.shape == (2, 2)):
            self.normal = np.array([(self.corners[:, 1] - self.corners[:, 0])[1], (-1)*(self.corners[:, 1] - self.corners[:, 0])[0]])
            self.dim = 2
        elif (self.corners.shape[0] == 3 and self.corners[0].shape[0] > 2):
            # compute the normal assuming the vertices are aranged counter
            # clock wise when the normal defines "up"
            i_min = np.argmin(self.corners[0,:])
            i_prev = i_min - 1 if i_min > 0 else self.corners.shape[1] - 1
            i_next = i_min + 1 if i_min < self.corners.shape[1] - 1 else 0
            self.normal = np.cross(self.corners[:, i_next] - self.corners[:, i_min], 
                                   self.corners[:, i_prev] - self.corners[:, i_min])
            self.dim = 3

            # Compute a basis for the plane and project the corners into that basis
            self.plane_basis = np.zeros((3,2), order='F', dtype=np.float32)
            localx = np.array(self.corners[:,1]-self.plane_point)
            self.plane_basis[:,0] = localx/np.linalg.norm(localx)
            localy = np.array(np.cross(self.normal, localx))
            self.plane_basis[:,1] = localy/np.linalg.norm(localy)
            self.corners_2d = np.concatenate((
                [ np.dot(self.corners.T - self.plane_point, self.plane_basis[:,0]) ], 
                [ np.dot(self.corners.T - self.plane_point, self.plane_basis[:,1]) ]
                ))
            self.corners_2d = np.array(self.corners_2d, order='F', dtype=np.float32)
        else:
            raise NameError('Wall.__init__ input error : corners must be an np.array dim 2x2 or 3xN, N>2')
        self.normal = self.normal/np.linalg.norm(self.normal)
        if (name is not None):
            self.name = name

    def intersection(self, p1, p2):
        '''
        Returns the intersection point between the wall and a line segment.
        
        :arg p1: (np.array dim 2 or 3) first end point of the line segment
        :arg p2: (np.array dim 2 or 3) second end point of the line segment
        
        :returns: (np.array dim 2 or 3 or None) intersection point between the wall and the line segment
        '''
        
        p1 = np.array(p1)
        p2 = np.array(p2)
    
        if (self.dim == 2):
            return geom.intersection_2D_segments(p1, p2, self.corners[:,0], self.corners[:,1])
            
        if (self.dim == 3):
            return geom.intersection_segment_polygon_surface(p1, p2, self.corners_2d, self.normal, self.plane_point, self.plane_basis)
        
    def intersects(self, p1, p2):
        '''
        Tests if the given line segment intersects the wall.
        
        :arg p1: (ndarray size 2 or 3) first endpoint of the line segment
        :arg p2: (ndarray size 2 or 3) second endpoint of the line segment
        
        :returns: (tuple size 3)
            (bool) True if the line segment intersects the wall
            (bool) True if the intersection happens at a border of the wall
            (bool) True if the intersection happens at the extremity of the segment
        '''
        
        if (self.dim == 2):
            intersection, borderOfSegment, borderOfWall = geom.intersection_2D_segments(p1, p2, self.corners[:,0], self.corners[:,1])

        if (self.dim == 3):
            intersection, borderOfSegment, borderOfWall = geom.intersection_segment_polygon_surface(p1, p2, self.corners_2d, self.normal,
                                                                                                 self.plane_point, self.plane_basis)

        if intersection is None:
            intersects = False
        else:
            intersects = True
            
        return intersects, borderOfWall, borderOfSegment
            
    def side(self, p):
        '''
        Computes on which side of the wall the point p is.
        
        :arg p: (np.array dim 2 or 3) coordinates of the point
        
        :returns: (int) integer representing on which side the point is
            -1 : opposite to the normal vector (going inside the room)
            0 : on the wall
            1 : in the direction of the normal vector (going outside of the room)
        '''
        
        p = np.array(p)
        if (self.dim != p.shape[0]):
            raise NameError('Wall.side input error : dimension of p and the wall must match.')

        return geom.side(p, self.corners[:,0], self.normal)




import numpy as np

def frac_delay(delta, N, w_max=0.9, C=4):
    '''
    Compute optimal fractionnal delay filter according to

    Design of Fractional Delay Filters Using Convex Optimization
    William Putnam and Julius Smith

    Parameters
    ----------
    delta: 
        delay of filter in (fractionnal) samples
    N: 
        number of taps
    w_max: 
        Bandwidth of the filter (in fraction of pi) (default 0.9)
    C: 
        sets the number of constraints to C*N (default 4)
    '''

    # constraints
    N_C = int(C*N)
    w = np.linspace(0, w_max*np.pi, N_C)[:,np.newaxis]
    
    n = np.arange(N)

    try:
        from cvxopt import solvers, matrix
    except:
        raise ValueError('To use the frac_delay function, the cvxopt module is necessary.')

    f = np.concatenate((np.zeros(N), np.ones(1)))

    A = []
    b = []
    for i in range(N_C):
        Anp = np.concatenate(([np.cos(w[i]*n), -np.sin(w[i]*n)], [[0],[0]]), axis=1)
        Anp = np.concatenate(([-f], Anp), axis=0)
        A.append(matrix(Anp))
        b.append(matrix(np.concatenate(([0], np.cos(w[i]*delta), -np.sin(w[i]*delta)))))

    solvers.options['show_progress'] = False
    sol = solvers.socp(matrix(f), Gq=A, hq=b)

    h = np.array(sol['x'])[:-1,0]

    '''
    import matplotlib.pyplot as plt
    w = np.linspace(0, np.pi, 2*N_C)
    F = np.exp(-1j*w[:,np.newaxis]*n)
    Hd = np.exp(-1j*delta*w)
    plt.figure()
    plt.subplot(3,1,1)
    plt.plot(np.abs(np.dot(F,h) - Hd))
    plt.subplot(3,1,2)
    plt.plot(np.diff(np.angle(np.dot(F,h))))
    plt.subplot(3,1,3)
    plt.plot(h)
    '''

    return h


def low_pass(numtaps, B, epsilon=0.1):

    bands = [0, (1-epsilon)*B, B, 0.5]
    desired = [1, 0]

    from scipy.signal import remez

    h = remez(numtaps, bands, desired, grid_density=32)

    '''
    import matplotlib.pyplot as plt
    w = np.linspace(0, np.pi, 8*numtaps)
    F = np.exp(-1j*w[:,np.newaxis]*np.arange(numtaps))
    Hd = np.exp(-1j*w)
    plt.figure()
    plt.subplot(3,1,1)
    plt.plot(np.abs(np.dot(F,h)))
    plt.subplot(3,1,2)
    plt.plot(np.angle(np.dot(F,h)))
    plt.subplot(3,1,3)
    plt.plot(h)
    '''
    
    return h

def resample(x, p, q):

    import fractions
    gcd = fractions.gcd(p,q)
    p /= gcd
    q /= gcd

    m = np.maximum(p,q)
    h = low_pass(10*m+1, 1./(2.*m))

    x_up = np.kron(x, np.concatenate(([1], np.zeros(p-1))))

    from scipy.signal import fftconvolve
    x_rs = fftconvolve(x_up, h)

    x_ds = x_rs[h.shape[0]/2+1::q]
    x_ds = x_ds[:np.floor(x.shape[0]*p/q)]

    '''
    import matplotlib.pyplot as plt
    plt.figure()
    plt.subplot(3,1,1)
    plt.plot(x)
    plt.subplot(3,1,2)
    plt.plot(x_rs)
    plt.subplot(3,1,3)
    plt.plot(x_ds)
    '''

    return x_ds




# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

import numpy as np

from .parameters import constants, eps


def area(corners):
    '''
    Computes the signed area of a 2D surface represented by its corners.
    
    :arg corners: (np.array 2xN, N>2) list of coordinates of the corners forming the surface
    
    :returns: (float) area of the surface
        positive area means anti-clockwise ordered corners.
        negative area means clockwise ordered corners.
    '''

    corners = np.array(corners)
    x = corners[0, :] - corners[0, range(-1, corners.shape[1]-1)]
    y = corners[1, :] + corners[1, range(-1, corners.shape[1]-1)]
    return -0.5 * (x * y).sum()


def side(p, p0, vector):
    '''
    Compute on which side of a given point an other given point is according to a vector.
    
    :arg p: (ndarray size 2 or 3) point to be tested
    :arg p0: (ndarray size 2 or 3) origin point
    :arg vector: (ndarray size 2 or 3) directional vector
    
    :returns: (int) direction of the point
        1 : p is at the side pointed by the vector
        0 : p is in the middle (on the same line as p0)
        -1 : p is at the opposite side of the one pointed by the vector
    '''
    p = np.array(p)
    p0 = np.array(p0)
    vector = np.array(vector)
    
    projection = np.dot(vector, p - p0)
    if (projection > 0):
        return 1
    elif (projection < 0):
        return -1
    else:
        return 0


def ccw3p(p1, p2, p3):
    '''
    Computes the orientation of three 2D points.
    
    :arg p1: (ndarray size 2) coordinates of a 2D point
    :arg p2: (ndarray size 2) coordinates of a 2D point
    :arg p3: (ndarray size 2) coordinates of a 2D point
    
    :returns: (int) orientation of the given triangle
        1 if triangle vertices are counter-clockwise
        -1 if triangle vertices are clockwise
        0 if vertices are collinear

    :ref: https://en.wikipedia.org/wiki/Curve_orientation
    '''
    
    p1, p2, p3 = np.array(p1), np.array(p2), np.array(p3)
    if (p1.shape[0] != 2 or p2.shape[0] != 2 or p3.shape[0] != 2):
        raise ValueError('geometry.ccw3p is for three 2D points')
    d = (p2[0] - p1[0]) * (p3[1] - p1[1]) - (p3[0] - p1[0]) * (p2[1] - p1[1])

    if (np.abs(d) < eps):
        return 0
    elif (d > 0):
        return 1
    else:
        return -1


def intersection_2D_segments(a1, a2, b1, b2) :
    '''
    Computes the intersection between two 2D line segments.
    
    This function computes the intersection between two 2D segments
    (defined by the coordinates of their endpoints) and returns the
    coordinates of the intersection point.
    If there is no intersection, None is returned.
    If segments are collinear, None is returned.
    Two booleans are also returned to indicate if the intersection
    happened at extremities of the segments, which can be useful for limit cases
    computations.
    
    :arg a1: (ndarray size 2) coordinates of the first endpoint of segment a
    :arg a2: (ndarray size 2) coordinates of the second endpoint of segment a
    :arg b1: (ndarray size 2) coordinates of the first endpoint of segment b
    :arg b2: (ndarray size 2) coordinates of the second endpoint of segment b
    
    :returns: (tuple of 3 elements) results of the computation
        (ndarray size 2 or None) coordinates of the intersection point
        (bool) True if the intersection is at boundaries of segment a
        (bool) True if the intersection is at boundaries of segment b
    '''
    
    a1=np.array(a1)
    a2=np.array(a2)
    b1=np.array(b1)
    b2=np.array(b2)
    if (a1.shape[0] != 2):
        raise NameError('utilities.intersection2DSegments input error : a1 is a ndarray of size 2')
    if (a2.shape[0] != 2):
        raise NameError('utilities.intersection2DSegments input error : a2 is a ndarray of size 2')
    if (b1.shape[0] != 2):
        raise NameError('utilities.intersection2DSegments input error : b1 is a ndarray of size 2')
    if (b2.shape[0] != 2):
        raise NameError('utilities.intersection2DSegments input error : b2 is a ndarray of size 2')

    # First we weed out the simple cases where no intersection happens
    # case 1
    a1a2b1 = ccw3p(a1, a2, b1)
    a1a2b2 = ccw3p(a1, a2, b2)
    if a1a2b1 == a1a2b2:
        return None, False, False

    # case 2
    b1b2a1 = ccw3p(b1, b2, a1)
    b1b2a2 = ccw3p(b1, b2, a2)
    if b1b2a1 == b1b2a2:
        return None, False, False

    da = a2-a1
    db = b2-b1
    dap = np.empty_like(da)
    dap[0] = -da[1]
    dap[1] = da[0]
    denom = np.dot(dap, db)

    # case 3
    if denom == 0:
        return None, False, False

    # At this point, we know there is intersection
    dp = a1-b1
    num = np.dot(dap, dp)
    # This is the intersection point
    p = np.array((num / denom.astype(float))*db + b1)

    # Test if intersection is actually at one of a1 or a2
    if b1b2a1 == 0 or b1b2a2 == 0:
        endpointA = True
    else:
        endpointA = False

    # Test if intersection is actually at one of b1 or b2
    if a1a2b1 == 0 or a1a2b2 == 0:
        endpointB = True
    else:
        endpointB = False

    return p, endpointA, endpointB


    
def intersection_segment_plane(a1, a2, p, normal):
    '''
    Computes the intersection between a line segment and a plane in 3D.
    
    This function computes the intersection between a line segment (defined
    by the coordinates of two points) and a plane (defined by a point belonging
    to it and a normal vector). If there is no intersection, None is returned.
    If the segment belongs to the surface, None is returned.
    A boolean is also returned to indicate if the intersection
    happened at extremities of the segment, which can be useful for limit cases
    computations.
    
    :arg a1: (ndarray size 3) coordinates of the first endpoint of the segment
    :arg a2: (ndarray size 3) coordinates of the second endpoint of the segment
    :arg p: (ndarray size 3) coordinates of a point belonging to the plane
    :arg normal: (ndarray size 3) normal vector of the plane
    
    :returns: (tuple of 2 elements) results of the computation
        (ndarray size 3 or None) coordinates of the intersection point
        (bool) True if the intersection is at boundaries of the segment
    '''

    a1=np.array(a1).astype(float)
    a2=np.array(a2).astype(float)
    p=np.array(p).astype(float)
    normal=np.array(normal).astype(float)
    if (a1.shape[0] != 3):
        raise NameError('utilities.intersectionSegmentPlane input error : a1 is a ndarray of size 3')
    if (a2.shape[0] != 3):
        raise NameError('utilities.intersectionSegmentPlane input error : a2 is a ndarray of size 3')
    if (p.shape[0] != 3):
        raise NameError('utilities.intersectionSegmentPlane input error : p is a ndarray of size 3')
    if (normal.shape[0] != 3):
        raise NameError('utilities.intersectionSegmentPlane input error : normal is a ndarray of size 3')
    
    u = a2-a1
    w = a1-p
    denom = np.dot(normal, u)

    if(abs(denom) < eps):
        return None, False

    else:
        num = -np.dot(normal, w)
        s = num/denom
        if(s<0 or s>1):
            return None, False
        else:
            if (s==0 or s==1):
                limitCase = True
            else:
                limitCase = False
            return a1+s*u, limitCase


def intersection_segment_polygon_surface(a1, a2, corners_2d, normal, plane_point, plane_basis):
    '''
    Computes the intersection between a line segment and a polygon surface in 3D.
    
    This function computes the intersection between a line segment (defined
    by the coordinates of two points) and a surface (defined by an array of
    coordinates of corners of the polygon and a normal vector)
    If there is no intersection, None is returned.
    If the segment belongs to the surface, None is returned.
    Two booleans are also returned to indicate if the intersection
    happened at extremities of the segment or at a border of the polygon,
    which can be useful for limit cases computations.
    
    :arg a1: (ndarray size 3) coordinates of the first endpoint of the segment
    :arg a2: (ndarray size 3) coordinates of the second endpoint of the segment
    :arg corners: (ndarray size 3xN, N>2) coordinates of the corners of the polygon
    :arg normal: (ndarray size 3) normal vector of the surface
    
    :returns: (tuple of 3 elements) results of the computation
        (ndarray size 3 or None) coordinates of the intersection point
        (bool) True if the intersection is at boundaries of the segment
        (bool) True if the intersection is at boundaries of the polygon
    '''

    a1 = np.array(a1)
    a2 = np.array(a2)
    corners_2d = np.array(corners_2d)
    normal = np.array(normal)
    if (a1.shape[0] != 3):
        raise NameError('utilities.intersectionSegmentPolygonSurface input error : a1 is a ndarray of size 3')
    if (a2.shape[0] != 3):
        raise NameError('utilities.intersectionSegmentPolygonSurface input error : a2 is a ndarray of size 3')
    if (corners_2d.shape[0] != 2 or corners_2d.shape[1] < 3):
        raise NameError('utilities.intersectionSegmentPolygonSurface input error : corners is a 3xN ndarray, N>2')
    if (normal.shape[0] != 3):
        raise NameError('utilities.intersectionSegmentPolygonSurface input error : normal is a ndarray of size 3')
    if (plane_point.shape[0] != 3):
        raise NameError('utilities.intersectionSegmentPolygonSurface input error : point_plane is a ndarray of size 3')
    if (plane_basis.shape != (3,2)):
        raise NameError('utilities.intersectionSegmentPolygonSurface input error : point_basis is a ndarray of shape (3,2)')
    
    # Check if the segment intersects the plane formed by the surface
    # Get the intersection point if it is the case
    p, segmentLimit = intersection_segment_plane(a1, a2, plane_point, normal)
    if p is None:
        return None, False, False

    # Project intersection in the plane basis
    localp = np.dot(plane_basis.T, p - plane_point)

    # Check if the intersection point is in the polygon on the plane
    inside, borderLimit = is_inside_2D_polygon(localp, corners_2d)
    if not inside:
        p = None
    
    return p, segmentLimit, borderLimit


def is_inside_2D_polygon(p, corners):
    '''
    Checks if a given point is inside a given polygon in 2D.
    
    This function checks if a point (defined by its coordinates) is inside
    a polygon (defined by an array of coordinates of its corners) by counting
    the number of intersections between the borders and a segment linking
    the given point with a computed point outside the polygon.
    A boolean is also returned to indicate if a point is on a border of the
    polygon (the point is still considered inside), which can be useful for
    limit cases computations.
    
    :arg p: (ndarray size 2) coordinates of the point
    :arg corners: (ndarray size 2xN, N>2) coordinates of the corners of the polygon
    
    :returns: (tuple of 2 elements) results of the computation
        (bool) True if the point is inside
        (bool) True if the intersection is at boundaries of the polygon
    '''
    
    p = np.array(p)
    corners = np.array(corners)
    if (p.shape[0] != 2):
        raise NameError('utilities.isInside2DPolygon input error : p is a ndarray of size 2')
    if (corners.shape[0] != 2 or corners.shape[1] < 3):
        raise NameError('utilities.isInside2DPolygon input error : corners is a 2xN ndarray, N>2')
        
    # Compute p0, which is a point outside the room at x coordinate xMin-1
    # (where xMin is the minimum x coordinate among the corners)
    p0 = np.array([np.amin(corners[0])-1, p[1]])

    is_inside = False

    j = corners.shape[1] - 1
    for i in range(corners.shape[1]):

        # Check first if the point is on the segment
        # We count the border as inside the polygon
        c1c2p = ccw3p(corners[:,i], corners[:,j], p)
        if c1c2p == 0:
            # Here we know that p is co-linear with the two corners
            x_down = min(corners[0,i], corners[0,j])
            x_up = max(corners[0,i], corners[0,j])
            y_down = min(corners[1,i], corners[1,j])
            y_up = max(corners[1,i], corners[1,j])
            if x_down <= p[0] and p[0] <= x_up and y_down <= p[1] and p[1] <= y_up:
                return True, True


        # Now check the intersection using standard algorithm
        c1c2p0 = ccw3p(corners[:,i], corners[:,j], p0)
        if c1c2p == c1c2p0:
            # we know there is no intersection
            j = i
            continue

        pp0c1 = ccw3p(p, p0, corners[:,i])
        pp0c2 = ccw3p(p, p0, corners[:,j])
        if pp0c1 == pp0c2:
            # we know there is no intersection
            j = i
            continue

        # At this point, we know there is an intersection

        # the second condition takes care of horizontal edges and intersection on vertex
        if p[1] < max(corners[1,i], corners[1,j]):
            is_inside = not is_inside

        # circular move around polygon
        j = i

    if is_inside:
        return True, False
    else:
        return False, False




# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

from __future__ import division

import numpy as np
import scipy.linalg as la

from .parameters import constants
from . import utilities as u
from .soundsource import build_rir_matrix
from . import windows
from . import stft


#=========================================================================
# Free (non-class-member) functions related to beamformer design
#=========================================================================


def H(A, **kwargs):
    '''Returns the conjugate (Hermitian) transpose of a matrix.'''

    return np.transpose(A, **kwargs).conj()


def sumcols(A):
    '''
    Sums the columns of a matrix (np.array).

    The output is a 2D np.array
    of dimensions M x 1.
    '''

    return np.sum(A, axis=1, keepdims=1)


def mdot(*args):
    '''Left-to-right associative matrix multiplication of multiple 2D ndarrays.'''

    ret = args[0]
    for a in args[1:]:
        ret = np.dot(ret, a)

    return ret


def distance(x, y):
    '''
    Computes the distance matrix E.

    E[i,j] = sqrt(sum((x[:,i]-y[:,j])**2)).
    x and y are DxN ndarray containing N D-dimensional vectors.
    '''

    # Assume x, y are arrays, *not* matrices
    x = np.array(x)
    y = np.array(y)

    # return np.sqrt((x[0,:,np.newaxis]-y[0,:])**2 + (x[1,:,np.newaxis]-y[1,:])**2)

    return np.sqrt(np.sum((x[:, :, np.newaxis] - y[:, np.newaxis, :])**2, axis=0))


def unit_vec2D(phi):
    return np.array([[np.cos(phi), np.sin(phi)]]).T


def linear_2D_array(center, M, phi, d):
    u = unit_vec2D(phi)
    return np.array(center)[:, np.newaxis] + d * \
        (np.arange(M)[np.newaxis, :] - (M - 1.) / 2.) * u


def circular_2D_array(center, M, phi0, radius):
    phi = np.arange(M) * 2. * np.pi / M
    return np.array(center)[:, np.newaxis] + radius * \
        np.vstack((np.cos(phi + phi0), np.sin(phi + phi0)))


def poisson_2D_array(center, M, d):
    ''' Create array of 2D positions drawn from Poisson process. '''

    from numpy.random import standard_exponential, randint

    R = d*standard_exponential((2, M))*(2*randint(0, 2, (2, M)) - 1)
    R = R.cumsum(axis=1)
    R -= R.mean(axis=1)[:, np.newaxis]
    R += np.array([center]).T

    return R


def square_2D_array(center, M, N, phi, d):

    c = linear_2D_array(center, M, phi+np.pi/2., d)
    R = np.zeros((2, M*N))
    for i in np.arange(M):
        R[:, i*N:(i+1)*N] = linear_2D_array(c[:, i], N, phi, d)

    return R

def spiral_2D_array(center, M, radius=1., divi=3, angle=None):
    '''
    generate microphone array layout randomly
    :param center: location of the center of the array
    :param radius: microphones are contained within a cirle of this radius (default 1)
    :param M: number of microphones
    :param divi: number of rotations of the spiral (default 3)
    :param angle: the angle offset of the spiral (default random)
    :return:
    '''
    num_seg = np.ceil(M / divi)

    pos_array_norm = np.linspace(0, radius, num=M, endpoint=False)

    pos_array_angle = np.reshape(np.tile(np.pi * 2 * np.arange(divi) / divi, num_seg),
                                 (divi, -1), order='F') + \
                      np.linspace(0, 2 * np.pi / divi,
                                  num=num_seg, endpoint=False)[np.newaxis, :]
    pos_array_angle = np.insert(pos_array_angle.flatten('F')[:M - 1], 0, 0)

    if angle is None:
        pos_array_angle += np.random.rand() * np.pi / divi
    else:
        pos_array_angle += angle

    pos_mic_x = pos_array_norm * np.cos(pos_array_angle)
    pos_mic_y = pos_array_norm * np.sin(pos_array_angle)

    return np.array([pos_mic_x, pos_mic_y])


def fir_approximation_ls(weights, T, n1, n2):

    freqs_plus = np.array(weights.keys())[:, np.newaxis]
    freqs = np.vstack([freqs_plus,
                       -freqs_plus])
    omega = 2 * np.pi * freqs
    omega_discrete = omega * T

    n = np.arange(n1, n2)

    # Create the DTFT transform matrix corresponding to a discrete set of
    # frequencies and the FIR filter indices
    F = np.exp(-1j * omega_discrete * n)

    w_plus = np.array(weights.values())[:, :, 0]
    w = np.vstack([w_plus,
                   w_plus.conj()])

    return np.linalg.pinv(F).dot(w)


#=========================================================================
# Classes (microphone array and beamformer related)
#=========================================================================


class MicrophoneArray(object):

    '''Microphone array class.'''

    def __init__(self, R, Fs):
        self.dim = R.shape[0]   # are we in 2D or in 3D
        self.M = R.shape[1]     # number of microphones
        self.R = R              # array geometry

        self.Fs = Fs            # sampling frequency of microphones

        self.signals = None

        self.center = np.mean(R, axis=1, keepdims=True)


    def record(self, signals, Fs):
        '''
        This simulates the recording of the signals by the microphones.
        In particular, if the microphones and the room simulation
        do not use the same sampling frequency, down/up-sampling
        is done here.

        Parameters:
        -----------

        signals: An ndarray with as many lines as there are microphones.
        Fs: the sampling frequency of the signals.
        '''

        if signals.shape[0] != self.M:
            raise NameError('The signals array should have as many lines as there are microphones.')

        if signals.ndim != 2:
            raise NameError('The signals should be a 2D array.')

        if Fs != self.Fs:
            from scikits.samplerate import resample
            Fs_ratio = self.Fs/float(Fs)
            newL = int(Fs_ratio*signals.shape[1])-1
            self.signals = np.zeros((self.M, newL))
            # scikits.samplerate resample function considers columns as channels (hence the transpose)
            for m in range(self.M):
                self.signals[m] = resample(signals[m], Fs_ratio, 'sinc_best')
        else:
            self.signals = signals


    def to_wav(self, filename, mono=False, norm=False, type=float):
        '''Save all the signals to wav files.'''
        from scipy.io import wavfile

        if mono is True:
            signal = self.signals[self.M/2]
        else:
            signal = self.signals.T  # each column is a channel

        if type is float:
            bits = None
        elif type is np.int8:
            bits = 8
        elif type is np.int16:
            bits = 16
        elif type is np.int32:
            bits = 32
        elif type is np.int64:
            bits = 64
        else:
            raise NameError('No such type.')

        if norm is True:
            from utilities import normalize
            signal = normalize(signal, bits=bits)

        signal = np.array(signal, dtype=type)

        wavfile.write(filename, self.Fs, signal)


class Beamformer(MicrophoneArray):

    '''
    Beamformer class.

    At some point, in some nice way, the design methods
    should also go here. Probably with generic arguments.
    '''

    def __init__(self, R, Fs, N=1024, Lg=None, hop=None, zpf=0, zpb=0):
        '''
        Parameters:
        -----------
        R        Mics positions
        Fs       Sampling frequency
        N=1024   Length of FFT, i.e. number of FD beamforming weights, equally spaced. Defaults to 1024.
        Lg=N     Length of time-domain filters. Default to N.
        hop=N/2  Hop length for frequency domain processing. Default to N/2.
        zpf=0    Front zero padding length for frequency domain processing. Default is 0.
        zpb=0    Zero padding length for frequency domain processing. Default is 0.

        '''
        MicrophoneArray.__init__(self, R, Fs)

        # only support even length (in freq)
        if N % 2 is 1:
            N += 1

        self.N = int(N)    # FFT length

        if Lg is None:
            self.Lg = N    # TD filters length
        else:
            self.Lg = int(Lg)

        # setup lengths for FD processing
        self.zpf = int(zpf)
        self.zpb = int(zpb)
        self.L = self.N - self.zpf - self.zpb
        if hop is None:
            self.hop = self.L/2
        else:
            self.hop = hop

        # for now only support equally spaced frequencies
        self.frequencies = np.arange(0, self.N/2+1)/float(self.N)*float(self.Fs)

        # weights will be computed later, the array is of shape (M, N/2+1)
        self.weights = None

        # the TD beamforming filters (M, Lg)
        self.filters = None

    def __add__(self, y):
        ''' Concatenates two beamformers together.'''

        newR = np.concatenate((self.R, y.R), axis=1)
        return Beamformer(newR, self.Fs, self.Lg, self.N, hop=self.hop, zpf=self.zpf, zpb=self.zpb)

    def filters_from_weights(self, non_causal=0.):
        ''' Compute time-domain filters from frequency domain weights.'''

        if self.weights is None:
            raise NameError('Weights must be defined.')

        self.filters = np.zeros((self.M, self.Lg))

        if self.N <= self.Lg:

            # go back to time domain and shift DC to center
            tw = np.fft.irfft(np.conj(self.weights), axis=1, n=self.N)
            self.filters[:, :self.N] = np.concatenate((tw[:, -self.N/2:], tw[:, :self.N/2]), axis=1)

        elif self.N > self.Lg:

            # Least-square projection
            for i in np.arange(self.M):
                Lgp = np.floor((1 - non_causal)*self.Lg)
                Lgm = self.Lg - Lgp
                # the beamforming weights in frequency are the complex conjugates of the FT of the filter
                w = np.concatenate((np.conj(self.weights[i]), self.weights[i, -2:0:-1]))

                # create partial Fourier matrix
                k = np.arange(self.N)[:, np.newaxis]
                l = np.concatenate((np.arange(self.N-Lgm, self.N), np.arange(Lgp)))
                F = np.exp(-2j*np.pi*k*l/float(self.N))

                self.filters[i] = np.real(np.linalg.lstsq(F, w)[0])

    def weights_from_filters(self):

        if self.filters is None:
            raise NameError('Filters must be defined.')

        # this is what we want to use, really.
        #self.weights = np.conj(np.fft.rfft(self.filters, n=self.N, axis=1))

        # quick hack to be able to use MKL acceleration package from anaconda
        self.weights = np.zeros((self.M, self.N/2+1), dtype=np.complex128)
        for m in range(self.M):
            self.weights[m] = np.conj(np.fft.rfft(self.filters[m], n=self.N))

    def steering_vector_2D(self, frequency, phi, dist, attn=False):

        phi = np.array([phi]).reshape(phi.size)

        # Assume phi and dist are measured from the array's center
        X = dist * np.array([np.cos(phi), np.sin(phi)]) + self.center

        D = distance(self.R, X)
        omega = 2 * np.pi * frequency

        if attn:
            # TO DO 1: This will mean slightly different absolute value for
            # every entry, even within the same steering vector. Perhaps a
            # better paradigm is far-field with phase carrier.
            return 1. / (4 * np.pi) / D * np.exp(-1j * omega * D / constants.get('c'))
        else:
            return np.exp(-1j * omega * D / constants.get('c'))

    def steering_vector_2D_from_point(self, frequency, source, attn=True, ff=False):
        ''' Creates a steering vector for a particular frequency and source

        Args:
            frequency
            source: location in cartesian coordinates
            attn: include attenuation factor if True
            ff:   uses far-field distance if true

        Return:
            A 2x1 ndarray containing the steering vector.
        '''

        X = np.array(source)
        if X.ndim == 1:
            X = source[:, np.newaxis]

        omega = 2 * np.pi * frequency

        # normalize for far-field if requested
        if (ff):
            # unit vectors pointing towards sources
            p = (X - self.center)
            p /= np.linalg.norm(p)

            # The projected microphone distances on the unit vectors
            D = np.dot(self.R.T, p)

            # subtract minimum in each column
            D -= np.min(D)

        else:

            D = distance(self.R, X)

        phase = np.exp(-1j * omega * D / constants.get('c'))

        if attn:
            # TO DO 1: This will mean slightly different absolute value for
            # every entry, even within the same steering vector. Perhaps a
            # better paradigm is far-field with phase carrier.
            return 1. / (4 * np.pi) / D * phase
        else:
            return phase

    def response(self, phi_list, frequency):

        i_freq = np.argmin(np.abs(self.frequencies - frequency))

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError('Beamforming weights or filters need to be computed first.')

        # For the moment assume that we are in 2D
        bfresp = np.dot(H(self.weights[:,i_freq]), self.steering_vector_2D(
            self.frequencies[i_freq], phi_list, constants.get('ffdist')))

        return self.frequencies[i_freq], bfresp

    def response_from_point(self, x, frequency):

        i_freq = np.argmin(np.abs(self.frequencies - frequency))

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError('Beamforming weights or filters need to be computed first.')

        # For the moment assume that we are in 2D
        bfresp = np.dot(H(self.weights[:, i_freq]), self.steering_vector_2D_from_point(
            self.frequencies[i_freq], x, attn=True, ff=False))

        return self.frequencies[i_freq], bfresp

    def plot_response_from_point(self, x, legend=None):

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError('Beamforming weights or filters need to be computed first.')

        if x.ndim == 0:
            x = np.array([x])

        import matplotlib.pyplot as plt

        HF = np.zeros((x.shape[1], self.frequencies.shape[0]), dtype=complex)
        for k, p in enumerate(x.T):
            for i, f in enumerate(self.frequencies):
                r = np.dot(H(self.weights[:, i]),
                           self.steering_vector_2D_from_point(f, p, attn=True, ff=False))
                HF[k, i] = r[0]

        plt.subplot(2, 1, 1)
        plt.title('Beamformer response')
        for hf in HF:
            plt.plot(self.frequencies, np.abs(hf))
        plt.ylabel('Modulus')
        plt.axis('tight')
        plt.legend(legend)

        plt.subplot(2, 1, 2)
        for hf in HF:
            plt.plot(self.frequencies, np.unwrap(np.angle(hf)))
        plt.ylabel('Phase')
        plt.xlabel('Frequency [Hz]')
        plt.axis('tight')
        plt.legend(legend)

    def plot_beam_response(self):

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError('Beamforming weights or filters need to be computed first.')

        phi = np.linspace(-np.pi, np.pi-np.pi/180, 360)
        freq = self.frequencies

        resp = np.zeros((freq.shape[0], phi.shape[0]), dtype=complex)

        for i, f in enumerate(freq):
            # For the moment assume that we are in 2D
            resp[i,:] = np.dot(H(self.weights[:,i]), self.steering_vector_2D(
                f, phi, constants.get('ffdist')))

        H_abs = np.abs(resp)**2
        H_abs /= H_abs.max()
        H_abs = 10*np.log10(H_abs)

        p_min = 0
        p_max = 100
        vmin, vmax = np.percentile(H_abs.flatten(), [p_min, p_max])

        import matplotlib.pyplot as plt

        plt.imshow(H_abs,
                   aspect='auto',
                   origin='lower',
                   interpolation='sinc',
                   vmax=vmax, vmin=vmin)

        plt.xlabel('Angle [rad]')
        xticks = [-np.pi, -np.pi/2, 0, np.pi/2, np.pi]
        for i, p in enumerate(xticks):
            xticks[i] = np.argmin(np.abs(p - phi))
        xticklabels = ['$-\pi$', '$-\pi/2$', '0', '$\pi/2$', '$\pi$']
        plt.setp(plt.gca(), 'xticks', xticks)
        plt.setp(plt.gca(), 'xticklabels', xticklabels)

        plt.ylabel('Freq [kHz]')
        yticks = np.zeros(4)
        f_0 = np.floor(self.Fs/8000.)
        for i in np.arange(1, 5):
            yticks[i-1] = np.argmin(np.abs(freq - 1000.*i*f_0))
        #yticks = np.array(plt.getp(plt.gca(), 'yticks'), dtype=np.int)
        plt.setp(plt.gca(), 'yticks', yticks)
        plt.setp(plt.gca(), 'yticklabels', np.arange(1, 5)*f_0)

    def snr(self, source, interferer, f, R_n=None, dB=False):

        i_f = np.argmin(np.abs(self.frequencies - f))

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError('Beamforming weights or filters need to be computed first.')

        # This works at a single frequency because otherwise we need to pass
        # many many covariance matrices. Easy to change though (you can also
        # have frequency independent R_n).

        if R_n is None:
            R_n = np.zeros((self.M, self.M))

        # To compute the SNR, we /must/ use the real steering vectors, so no
        # far field, and attn=True
        A_good = self.steering_vector_2D_from_point(self.frequencies[i_f], source.images, attn=True, ff=False)

        if interferer is not None:
            A_bad  = self.steering_vector_2D_from_point(self.frequencies[i_f], interferer.images, attn=True, ff=False)
            R_nq = R_n + sumcols(A_bad) * H(sumcols(A_bad))
        else:
            R_nq = R_n

        w = self.weights[:, i_f]
        a_1 = sumcols(A_good)

        SNR = np.real(mdot(H(w), a_1, H(a_1), w) / mdot(H(w), R_nq, w))

        if dB is True:
            SNR = 10 * np.log10(SNR)

        return SNR

    def udr(self, source, interferer, f, R_n=None, dB=False):

        i_f = np.argmin(np.abs(self.frequencies - f))

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is None and self.filters is None:
            raise NameError('Beamforming weights or filters need to be computed first.')

        if R_n is None:
            R_n = np.zeros((self.M, self.M))

        A_good = self.steering_vector_2D_from_point(self.frequencies[i_f], source.images, attn=True, ff=False)

        if interferer is not None:
            A_bad  = self.steering_vector_2D_from_point(self.frequencies[i_f], interferer.images, attn=True, ff=False)
            R_nq = R_n + sumcols(A_bad).dot(H(sumcols(A_bad)))
        else:
            R_nq = R_n

        w = self.weights[:, i_f]

        UDR = np.real(mdot(H(w), A_good, H(A_good), w) / mdot(H(w), R_nq, w))
        if dB is True:
            UDR = 10 * np.log10(UDR)

        return UDR

    def process(self, FD=False):

        if self.signals is None or len(self.signals) == 0:
            raise NameError('No signal to beamform')

        if FD is True:

            # STFT processing

            if self.weights is None and self.filters is not None:
                self.weights_from_filters()
            elif self.weights is None and self.filters is None:
                raise NameError('Beamforming weights or filters need to be computed first.')

            # create window function
            win = np.concatenate((np.zeros(self.zpf),
                                  windows.hann(self.L),
                                  np.zeros(self.zpb)))

            # do real STFT of first signal
            tfd_sig = stft.stft(self.signals[0],
                                self.L,
                                self.hop,
                                zp_back=self.zpb,
                                zp_front=self.zpf,
                                transform=np.fft.rfft,
                                win=win) * np.conj(self.weights[0])
            for i in range(1, self.M):
                tfd_sig += stft.stft(self.signals[i],
                                     self.L,
                                     self.hop,
                                     zp_back=self.zpb,
                                     zp_front=self.zpf,
                                     transform=np.fft.rfft,
                                     win=win) * np.conj(self.weights[i])

            #  now reconstruct the signal
            output = stft.istft(
                tfd_sig,
                self.L,
                self.hop,
                zp_back=self.zpb,
                zp_front=self.zpf,
                transform=np.fft.irfft)

            # remove the zero padding from output signal
            if self.zpb is 0:
                output = output[self.zpf:]
            else:
                output = output[self.zpf:-self.zpb]

        else:

            # TD processing

            if self.weights is not None and self.filters is None:
                self.filters_from_weights()
            elif self.weights is None and self.filters is None:
                raise NameError('Beamforming weights or filters need to be computed first.')

            from scipy.signal import fftconvolve

            # do real STFT of first signal
            output = fftconvolve(self.filters[0], self.signals[0])
            for i in range(1, len(self.signals)):
                output += fftconvolve(self.filters[i], self.signals[i])

        return output

    def plot(self, sum_ir=False, FD=True):

        if self.weights is None and self.filters is not None:
            self.weights_from_filters()
        elif self.weights is not None and self.filters is None:
            self.filters_from_weights()
        elif self.weights is None and self.filters is None:
            raise NameError('Beamforming weights or filters need to be computed first.')

        import matplotlib.pyplot as plt

        if FD is True:
            plt.subplot(2, 2, 1)
            plt.plot(self.frequencies, np.abs(self.weights.T))
            plt.title('Beamforming weights [modulus]')
            plt.xlabel('Frequency [Hz]')
            plt.ylabel('Weight modulus')

            plt.subplot(2, 2, 2)
            plt.plot(self.frequencies, np.unwrap(np.angle(self.weights.T), axis=0))
            plt.title('Beamforming weights [phase]')
            plt.xlabel('Frequency [Hz]')
            plt.ylabel('Unwrapped phase')

            plt.subplot(2, 1, 2)

        plt.plot(np.arange(self.Lg)/float(self.Fs), self.filters.T)

        plt.title('Beamforming filters')
        plt.xlabel('Time [s]')
        plt.ylabel('Filter amplitude')
        plt.axis('tight')

    def far_field_weights(self, phi):
        '''
        This method computes weight for a far field at infinity

        phi: direction of beam
        '''

        u = unit_vec2D(phi)
        proj = np.dot(u.T, self.R - self.center)[0]

        # normalize the first arriving signal to ensure a causal filter
        proj -= proj.max()

        self.weights = np.exp(2j * np.pi * 
        self.frequencies[:, np.newaxis] * proj / constants.get('c')).T


    def rake_delay_and_sum_weights(self, source, interferer=None, R_n=None, attn=True, ff=False):

        self.weights = np.zeros((self.M, self.frequencies.shape[0]), dtype=complex)

        K = source.images.shape[1] - 1

        for i, f in enumerate(self.frequencies):
            W = self.steering_vector_2D_from_point(f, source.images, attn=attn, ff=ff)
            self.weights[:,i] = 1.0/self.M/(K+1) * np.sum(W, axis=1)


    def rake_one_forcing_weights(self, source, interferer, R_n=None, ff=False, attn=True):

        if R_n is None:
            R_n = np.zeros((self.M, self.M))

        self.weights = np.zeros((self.M, self.frequencies.shape[0]), dtype=complex)

        for i, f in enumerate(self.frequencies):
            if interferer is None:
                A_bad = np.array([[]])
            else:
                A_bad = self.steering_vector_2D_from_point(f, interferer.images, attn=attn, ff=ff)

            R_nq = R_n + sumcols(A_bad).dot(H(sumcols(A_bad)))

            A_s      = self.steering_vector_2D_from_point(f, source.images, attn=attn, ff=ff)
            R_nq_inv = np.linalg.pinv(R_nq)
            D = np.linalg.pinv(mdot(H(A_s), R_nq_inv, A_s))

            self.weights[:, i] = sumcols(mdot(R_nq_inv, A_s, D))[:, 0]

    def rake_max_sinr_weights(self, source, interferer, R_n=None,
                           rcond=0., ff=False, attn=True):
        '''
        This method computes a beamformer focusing on a number of specific sources
        and ignoring a number of interferers.

        INPUTS
          * source     : source locations
          * interferer : interferer locations
        '''

        if R_n is None:
            R_n = np.zeros((self.M, self.M))

        self.weights = np.zeros((self.M, self.frequencies.shape[0]), dtype=complex)

        for i, f in enumerate(self.frequencies):

            A_good = self.steering_vector_2D_from_point(f, source.images, attn=attn, ff=ff)

            if interferer is None:
                A_bad = np.array([[]])
            else:
                A_bad = self.steering_vector_2D_from_point(f, interferer.images, attn=attn, ff=ff)

            a_good = sumcols(A_good)
            a_bad = sumcols(A_bad)

            # TO DO: Fix this (check for numerical rank, use the low rank approximation)
            K_inv = np.linalg.pinv(a_bad.dot(H(a_bad)) + R_n + rcond * np.eye(A_bad.shape[0]))
            self.weights[:, i] = (K_inv.dot(a_good) / mdot(H(a_good), K_inv, a_good))[:, 0]

    def rake_max_udr_weights(self, source, interferer, R_n=None, ff=False, attn=True):
        
        if source.images.shape[1] == 1:
            self.rake_max_sinr_weights(source.images, interferer.images, R_n=R_n, ff=ff, attn=attn)
            return

        if R_n is None:
            R_n = np.zeros((self.M, self.M))

        self.weights = np.zeros((self.M, self.frequencies.shape[0]), dtype=complex)

        for i, f in enumerate(self.frequencies):
            A_good = self.steering_vector_2D_from_point(f, source.images, attn=attn, ff=ff)

            if interferer is None:
                A_bad = np.array([[]])
            else:
                A_bad = self.steering_vector_2D_from_point(f, interferer.images, attn=attn, ff=ff)

            R_nq = R_n + sumcols(A_bad).dot(H(sumcols(A_bad)))

            C = np.linalg.cholesky(R_nq)
            l, v = np.linalg.eig(mdot(np.linalg.inv(C), A_good, H(A_good), H(np.linalg.inv(C))))

            self.weights[:, i] = np.linalg.inv(H(C)).dot(v[:, 0])

    def rake_max_udr_filters(self, source, interferer, R_n, delay=0.03, epsilon=5e-3):
        '''
        Compute directly the time-domain filters for a UDR maximizing beamformer.
        '''

        H = build_rir_matrix(self.R, (source, interferer), self.Lg, self.Fs, epsilon=epsilon, unit_damping=True)
        L = H.shape[1]/2

        # Delay of the system in samples
        kappa = int(delay*self.Fs)
        precedence = int(0.030*self.Fs)

        # the constraint
        n = np.minimum(L, kappa+precedence)
        Hnc = H[:, :kappa]
        Hpr = H[:, kappa:n]
        Hc = H[:, n:L]
        A = np.dot(Hpr, Hpr.T)
        B = np.dot(Hnc, Hnc.T) + np.dot(Hc, Hc.T) + np.dot(H[:, L:], H[:, L:].T) + R_n

        # solve the problem
        SINR, v = la.eigh(A, b=B, eigvals=(self.M*self.Lg-1, self.M*self.Lg-1), overwrite_a=True, overwrite_b=True, check_finite=False)
        g_val = np.real(v[:, 0])

        # reshape and store
        self.filters = g_val.reshape((self.M, self.Lg))

        # compute and return SNR
        return SINR[0]


    def rake_perceptual_filters(self, source, interferer, R_n, delay=0.03, d_relax=0.035, epsilon=5e-3):
        '''
        Compute directly the time-domain filters for a perceptually motivated beamformer.
        The beamformer minimizes noise and interference, but relaxes the response of the
        filter within the 30 ms following the delay.
        '''

        # build the channel matrix
        H = build_rir_matrix(self.R, (source, interferer), self.Lg, self.Fs, epsilon=epsilon, unit_damping=True)
        L = H.shape[1] // 2

        # Delay of the system in samples
        tau = int(delay*self.Fs)
        kappa = int(d_relax*self.Fs)

        # the constraint
        A = np.concatenate((H[:,:tau+1], H[:,tau+kappa:]), axis=1)
        b = np.zeros((A.shape[1],1))
        b[tau,0] = 1

        # We first assume the sample are uncorrelated
        K_nq = np.dot(H[:, L:], H[:, L:].T) + R_n

        # causal response construction
        C = la.cho_factor(K_nq, overwrite_a=True, check_finite=False)
        B = la.cho_solve(C, A)
        D = np.dot(A.T, B)
        C = la.cho_factor(D, overwrite_a=True, check_finite=False)
        x = la.cho_solve(C, b)
        g_val = np.dot(B, x)

        # reshape and store
        self.filters = g_val.reshape((self.M, self.Lg))

        # compute and return SNR
        A = np.dot(g_val.T, H[:, :L])
        num = np.dot(A, A.T)
        denom =  np.dot(np.dot(g_val.T, K_nq), g_val)

        return num/denom


    def rake_max_sinr_filters(self, source, interferer, R_n, epsilon=5e-3, delay=0.):
        '''
        Compute the time-domain filters of SINR maximizing beamformer.
        '''

        H = build_rir_matrix(self.R, (source, interferer), self.Lg, self.Fs, epsilon=epsilon, unit_damping=True)
        L = H.shape[1]/2

        # We first assume the sample are uncorrelated
        K_s = np.dot(H[:, :L], H[:, :L].T)
        K_nq = np.dot(H[:, L:], H[:, L:].T) + R_n

        # Compute TD filters using generalized Rayleigh coefficient maximization
        SINR, v = la.eigh(K_s, b=K_nq, eigvals=(self.M*self.Lg-1, self.M*self.Lg-1), overwrite_a=True, overwrite_b=True, check_finite=False)
        g_val = np.real(v[:, 0])

        self.filters = g_val.reshape((self.M, self.Lg))

        # compute and return SNR
        return SINR[0]

    def rake_distortionless_filters(self, source, interferer, R_n, delay=0.03, epsilon=5e-3):
        '''
        Compute time-domain filters of a beamformer minimizing noise and interference
        while forcing a distortionless response towards the source.
        '''

        H = build_rir_matrix(self.R, (source, interferer), self.Lg, self.Fs, epsilon=epsilon, unit_damping=True)
        L = H.shape[1]/2

        # We first assume the sample are uncorrelated
        K_nq = np.dot(H[:, L:], H[:, L:].T) + R_n

        # constraint
        kappa = int(delay*self.Fs)
        A = H[:, :L]
        b = np.zeros((L, 1))
        b[kappa, 0] = 1

        # filter computation
        C = la.cho_factor(K_nq, overwrite_a=True, check_finite=False)
        B = la.cho_solve(C, A)
        D = np.dot(A.T, B)
        C = la.cho_factor(D, overwrite_a=True, check_finite=False)
        x = la.cho_solve(C, b)
        g_val = np.dot(B, x)

        # reshape and store
        self.filters = g_val.reshape((self.M, self.Lg))

        # compute and return SNR
        A = np.dot(g_val.T, H[:, :L])
        num = np.dot(A, A.T)
        denom = np.dot(np.dot(g_val.T, K_nq), g_val)

        return num/denom

    def rake_mvdr_filters(self, source, interferer, R_n, delay=0.03, epsilon=5e-3):
        '''
        Compute the time-domain filters of the minimum variance distortionless
        response beamformer.
        '''

        H = build_rir_matrix(self.R, (source, interferer), self.Lg, self.Fs, epsilon=epsilon, unit_damping=True)
        L = H.shape[1] // 2

        # the constraint vector
        kappa = int(delay*self.Fs)
        h = H[:, kappa]

        # We first assume the sample are uncorrelated
        R_xx = np.dot(H[:, :L], H[:, :L].T)
        K_nq = np.dot(H[:, L:], H[:, L:].T) + R_n

        # Compute the TD filters
        C = la.cho_factor(R_xx + K_nq, check_finite=False)
        g_val = la.cho_solve(C, h)

        g_val /= np.inner(h, g_val)
        self.filters = g_val.reshape((self.M, self.Lg))

        # compute and return SNR
        num = np.inner(g_val.T, np.dot(R_xx, g_val))
        denom = np.inner(np.dot(g_val.T, K_nq), g_val)

        return num/denom

    def rake_one_forcing_filters(self, sources, interferers, R_n, epsilon=5e-3):
        '''
        Compute the time-domain filters of a beamformer with unit response
        towards multiple sources.
        '''

        dist_mat = distance(self.R, sources.images)
        s_time = dist_mat / constants.get('c')
        s_dmp = 1./(4*np.pi*dist_mat)

        dist_mat = distance(self.R, interferers.images)
        i_time = dist_mat / constants.get('c')
        i_dmp = 1./(4*np.pi*dist_mat)

        # compute offset needed for decay of sinc by epsilon
        offset = np.maximum(s_dmp.max(), i_dmp.max())/(np.pi*self.Fs*epsilon)
        t_min = np.minimum(s_time.min(), i_time.min())
        t_max = np.maximum(s_time.max(), i_time.max())

        # adjust timing
        s_time -= t_min - offset
        i_time -= t_min - offset
        Lh = np.ceil((t_max - t_min + 2*offset)*float(self.Fs))

        # the channel matrix
        K = sources.images.shape[1]
        Lg = self.Lg
        off = (Lg - Lh)/2
        L = self.Lg + Lh - 1

        H = np.zeros((Lg*self.M, 2*L))
        As = np.zeros((Lg*self.M, K))

        for r in np.arange(self.M):

            # build constraint matrix
            hs = u.low_pass_dirac(s_time[r, :, np.newaxis], s_dmp[r, :, np.newaxis], self.Fs, Lh)[:, ::-1]
            As[r*Lg+off:r*Lg+Lh+off, :] = hs.T

            # build interferer RIR matrix
            hx = u.low_pass_dirac(s_time[r, :, np.newaxis], s_dmp[r, :, np.newaxis], self.Fs, Lh).sum(axis=0)
            H[r*Lg:(r+1)*Lg, :L] = u.convmtx(hx, Lg).T

            # build interferer RIR matrix
            hq = u.low_pass_dirac(i_time[r, :, np.newaxis], i_dmp[r, :, np.newaxis], self.Fs, Lh).sum(axis=0)
            H[r*Lg:(r+1)*Lg, L:] = u.convmtx(hq, Lg).T

        ones = np.ones((K, 1))

        # We first assume the sample are uncorrelated
        K_x = np.dot(H[:, :L], H[:, :L].T)
        K_nq = np.dot(H[:, L:], H[:, L:].T) + R_n

        # Compute the TD filters
        K_nq_inv = np.linalg.inv(K_x+K_nq)
        C = np.dot(K_nq_inv, As)
        B = np.linalg.inv(np.dot(As.T, C))
        g_val = np.dot(C, np.dot(B, ones))
        self.filters = g_val.reshape((self.M, Lg))

        # compute and return SNR
        A = np.dot(g_val.T, H[:, :L])
        num = np.dot(A, A.T)
        denom = np.dot(np.dot(g_val.T, K_nq), g_val)

        return num/denom



# @version: 1.0  date: 09/07/2015 by Robin Scheibler
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

'''
This file defines the main physical constants of the system
'''

# tolerance for computations
eps = 1e-10

# We implement the constants as a dictionnary so that they can
# be modified at runtime.
# The class Constants gives an interface to update the value of
# constants or add new ones.
_constants = {}
_constants_default = { 
    'c' : 343.0,      # speed of sound at 20 C in dry air
    'ffdist' : 10.,   # distance to the far field
    'fc_hp' : 300.,   # cut-off frequency of standard high-pass filter
    'frac_delay_length' : 81, # Length of the fractional delay filters used for RIR gen
    }


class Constants:
    '''
    A class to provide easy access package wide to user settable constants.
    
    Be careful of not using this in tight loops since it uses exceptions.
    '''

    def set(self, name, val):
        # add constant to dictionnary
        _constants[name] = val

    def get(self, name):

        try:
            v = _constants[name]
        except KeyError: 
            try:
                v = _constants_default[name]
            except KeyError:
                raise NameError(name + ': no such constant')

        return v


# the instanciation of the class
constants = Constants()

# Compute the speed of sound as a function 
# of temperature, humidity, and pressure
def calculate_speed_of_sound(t, h, p):
    '''
    Compute the speed of sound as a function of
    temperature, humidity and pressure

    Parameters
    ----------
    t: 
        temperature [Celsius]
    h: 
        relative humidity [%]
    p: 
        atmospheric pressure [kpa]

    Returns
    -------

    Speed of sound in [m/s]
    '''

    # using crude approximation for now
    return 331.4 + 0.6*t + 0.0124*h




#!/usr/bin/env python
from __future__ import print_function

from pyroomacoustics import __version__

try:
    from setuptools import setup
    from setuptools import Extension
except ImportError:
    print("import error")
    from distutils.core import setup
    from distutils.extension import Extension

# To use a consistent encoding
from codecs import open
from os import path

# build C extension for image source model
src_dir = 'pyroomacoustics/c_package/'
files = ['wall.c', 'linalg.c', 'room.c', 'is_list.c', 'shoebox.c']

libroom_ext = Extension('pyroomacoustics.c_package.libroom',
                    extra_compile_args = ['-Wall', '-O3', '-std=c99'],
                    sources = [src_dir + f for f in files],
                    include_dirs=[src_dir])

libroom_lib = ('pyroomacoustics.c_package.libroom', {
    'sources': [src_dir + f for f in files],
    'extra_compile_args': ['-Wall', '-O3', '-std=c99'],
    'include_dirs':[src_dir]
    })

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup_kwargs = dict(
        name='pyroomacoustics',

        version=__version__,

        description='A simple framework for room acoustics and signal processing in Python.',
        long_description=long_description,

        author='Laboratory for Audiovisual Communications, EPFL',

        author_email='fakufaku@gmail.ch',

        url='https://github.com/LCAV/pyroomacoustics',

        license='GPL3',

        # You can just specify the packages manually here if your project is
        # simple. Or you can use find_packages().
        packages=['pyroomacoustics', 'pyroomacoustics.c_package'],

        # Libroom C extension
        ext_modules=[libroom_ext],

        install_requires=[
            'numpy',
            'scipy',
            ],

        test_suite='nose.collector',
        tests_require=['nose'],

        classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 4 - Beta',

            # Indicate who your project is intended for
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Build Tools',

            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            #'Programming Language :: Python :: 3.3',
            #'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            ],

        # What does your project relate to?
        keywords='room acoustics signal processing',
)

try:
    # Try to build everything first
    setup(**setup_kwargs)

except:
    # Retry without the C module
    print("Error. Probably building C extension failed. Retrying without.")
    setup_kwargs.pop('ext_modules')
    setup(**setup_kwargs)
    




# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

from __future__ import print_function

import numpy as np
import scipy.spatial as spatial
import ctypes

#import .beamforming as bf
from . import beamforming as bf
from . import geometry as geom
from .soundsource import SoundSource
from .wall import Wall
from .geometry import area, ccw3p
from .parameters import constants, eps

from .c_package import libroom_available, CWALL, CROOM, libroom, c_wall_p, c_int_p, c_float_p, c_room_p

class Room(object):
    '''
    This class represents a room instance.
    
    A room instance is formed by wall instances. A MicrophoneArray and SoundSources can be added.
    
    :attribute walls: (Wall array) list of walls forming the room
    :attribute fs: (int) sampling frequency
    :attribute t0: (float) time offset
    :attribute max_order: (int) the maximum computed order for images
    :attribute sigma2_awgn: (float) ambient additive white gaussian noise level
    :attribute sources: (SoundSource array) list of sound sources
    :attribute mics: (MicrophoneArray) array of microphones
    :attribute normals: (numpy.ndarray 2xN or 3xN, N=number of walls) array containing normal vector for each wall, used for calculations
    :attribute corners: (numpy.ndarray 2xN or 3xN, N=number of walls) array containing a point belonging to each wall, used for calculations
    :attribute absorption: (numpy.ndarray size N, N=number of walls)  array containing the absorption factor for each wall, used for calculations
    :attribute dim: (int) dimension of the room (2 or 3 meaning 2D or 3D)
    :attribute wallsId: (int dictionary) stores the mapping "wall name -> wall id (in the array walls)"
    '''

    def __init__(
            self,
            walls,
            fs=8000,
            t0=0.,
            max_order=1,
            sigma2_awgn=None,
            sources=None,
            mics=None):

        self.walls = walls
        self.fs = fs
        self.max_order = max_order
        self.sigma2_awgn = sigma2_awgn

        # Compute the filter delay if not provided
        if t0 < (constants.get('frac_delay_length')-1)/float(fs)/2:
            self.t0 = (constants.get('frac_delay_length')-1)/float(fs)/2
        else:
            self.t0 = t0
        
        if (sources is list):
            self.sources = sources
        else:
            self.sources = []

        self.mic_array = mics
         
        self.normals = np.array([wall.normal for wall in self.walls]).T
        self.corners = np.array([wall.corners[:, 0] for wall in self.walls]).T
        self.absorption = np.array([wall.absorption for wall in self.walls])

        # Pre-compute RIR if needed
        if (len(self.sources) > 0 and self.mic_array is not None):
            self.compute_rir()
        else:
            self.rir = None

        # in the beginning, nothing has been 
        self.visibility = None

        # Get the room dimension from that of the walls
        self.dim = walls[0].dim

        # mapping between wall names and indices
        self.wallsId = {}
        for i in range(len(walls)):
            if self.walls[i].name is not None:
                self.wallsId[self.walls[i].name] = i

        # check which walls are part of the convex hull
        self.convex_hull()

    @classmethod
    def from_corners(
            cls,
            corners,
            absorption=0.,
            fs=8000,
            t0=0.,
            max_order=1,
            sigma2_awgn=None,
            sources=None,
            mics=None):
        '''
        Creates a 2D room by giving an array of corners.
        
        :arg corners: (np.array dim 2xN, N>2) list of corners, must be antiClockwise oriented
        :arg absorption: (float array or float) list of absorption factor for each wall or single value for all walls
        
        :returns: (Room) instance of a 2D room
        '''
        
        corners = np.array(corners)
        if (corners.shape[0] != 2 or corners.shape[1] < 3):
            raise ValueError('Arg corners must be more than two 2D points.')

        if (geom.area(corners) <= 0):
            corners = corners[:,::-1]

        cls.corners = corners
        cls.dim = corners.shape[0] 
            
        absorption = np.array(absorption, dtype='float64')
        if (absorption.ndim == 0):
            absorption = absorption * np.ones(corners.shape[1])
        elif (absorption.ndim > 1 and corners.shape[1] != len(absorption)):
            raise ValueError('Arg absorption must be the same size as corners or must be a single value.')
        
        walls = []
        for i in range(corners.shape[1]):
            walls.append(Wall(np.array([corners[:, i], corners[:, (i+1)%corners.shape[1]]]).T, absorption[i], "wall_"+str(i)))
            
        return cls(walls, fs, t0, max_order, sigma2_awgn, sources, mics)

    def extrude(
            self,
            height,
            v_vec=None,
            absorption=0.):
        '''
        Creates a 3D room by extruding a 2D polygon. 
        The polygon is typically the floor of the room and will have z-coordinate zero. The ceiling

        Parameters
        ----------
        height : float
            The extrusion height
        v_vec : array-like 1D length 3, optionnal
            A unit vector. An orientation for the extrusion direction. The
            ceiling will be placed as a translation of the floor with respect
            to this vector (The default is [0,0,1]).
        absorption : float or array-like
            Absorption coefficients for all the walls. If a scalar, then all the walls
            will have the same absorption. If an array is given, it should have as many elements
            as there will be walls, that is the number of vertices of the polygon plus two. The two
            last elements are for the floor and the ceiling, respectively. (default 1)
        '''

        if self.dim != 2:
            raise ValueError('Can only extrude a 2D room.')

        # default orientation vector is pointing up
        if v_vec is None:
            v_vec = np.array([0., 0., 1.])

        # check that the walls are ordered counterclock wise
        # that should be the case if created from from_corners function
        nw = len(self.walls)
        floor_corners = np.zeros((2,nw))
        floor_corners[:,0] = self.walls[0].corners[:,0]
        ordered = True
        for iw, wall in enumerate(self.walls[1:]):
            if not np.allclose(self.walls[iw].corners[:,1], wall.corners[:,0]):
                ordered = False
            floor_corners[:,iw+1] = wall.corners[:,0]
        if not np.allclose(self.walls[-1].corners[:,1], self.walls[0].corners[:,0]):
            ordered = False

        if not ordered:
            raise ValueError("The wall list should be ordered counter-clockwise, which is the case \
                if the room is created with Room.from_corners")

        # make sure the floor_corners are ordered anti-clockwise (for now)
        if (geom.area(floor_corners) <= 0):
            floor_corners = np.fliplr(floor_corners)

        walls = []
        for i in range(nw):
            corners = np.array([
                np.r_[floor_corners[:,i], 0],
                np.r_[floor_corners[:,(i+1)%nw], 0],
                np.r_[floor_corners[:,(i+1)%nw], 0] + height*v_vec,
                np.r_[floor_corners[:,i], 0] + height*v_vec
                ]).T
            walls.append(Wall(corners, self.walls[i].absorption, name=str(i)))

        absorption = np.array(absorption)
        if absorption.ndim == 0:
            absorption = absorption * np.ones(2)
        elif absorption.ndim == 1 and absorption.shape[0] != 2:
            raise ValueError("The size of the absorption array must be 2 for extrude, for the floor and ceiling")

        floor_corners = np.pad(floor_corners, ((0, 1),(0,0)), mode='constant')
        ceiling_corners = (floor_corners.T + height*v_vec).T

        # we need the floor corners to ordered clockwise (for the normal to point outward)
        floor_corners = np.fliplr(floor_corners)

        walls.append(Wall(floor_corners, absorption[0], name='floor'))
        walls.append(Wall(ceiling_corners, absorption[1], name='ceiling'))

        self.walls = walls
        self.dim = 3

        # re-collect all normals, corners, absoption
        self.normals = np.array([wall.normal for wall in self.walls]).T
        self.corners = np.array([wall.corners[:, 0] for wall in self.walls]).T
        self.absorption = np.array([wall.absorption for wall in self.walls])

        # recheck which walls are in the convex hull
        self.convex_hull()


    def convex_hull(self):

        ''' 
        Finds the walls that are not in the convex hull
        '''

        all_corners = []
        for wall in self.walls[1:]:
            all_corners.append(wall.corners.T)
        X = np.concatenate(all_corners, axis=0)
        convex_hull = spatial.ConvexHull(X, incremental=True)

        # Now we need to check which walls are on the surface
        # of the hull
        self.in_convex_hull = [False] * len(self.walls)
        for i, wall in enumerate(self.walls):
            # We check if the center of the wall is co-linear or co-planar
            # with a face of the convex hull
            point = np.mean(wall.corners, axis=1)

            for simplex in convex_hull.simplices:
                if point.shape[0] == 2:
                    # check if co-linear
                    p0 = convex_hull.points[simplex[0]]
                    p1 = convex_hull.points[simplex[1]]
                    if geom.ccw3p(p0, p1, point) == 0:
                        # co-linear point add to hull
                        self.in_convex_hull[i] = True

                elif point.shape[0] == 3:
                    # Check if co-planar
                    p0 = convex_hull.points[simplex[0]]
                    p1 = convex_hull.points[simplex[1]]
                    p2 = convex_hull.points[simplex[2]]

                    normal = np.cross(p1 - p0, p2 - p0)
                    if np.abs(np.inner(normal, point - p0)) < eps:
                        # co-planar point found!
                        self.in_convex_hull[i] = True

        self.obstructing_walls = np.array([i for i in range(len(self.walls)) if not self.in_convex_hull[i]], dtype=np.int32)


    def plot(self, img_order=None, freq=None, figsize=None, no_axis=False, mic_marker_size=10, **kwargs):
        ''' Plots the room with its walls, microphones, sources and images '''
    
        import matplotlib
        from matplotlib.patches import Circle, Wedge, Polygon
        from matplotlib.collections import PatchCollection
        import matplotlib.pyplot as plt

        if (self.dim == 2):
            fig = plt.figure(figsize=figsize)

            if no_axis is True:
                ax = fig.add_axes([0, 0, 1, 1], aspect='equal', **kwargs)
                ax.axis('off')
                rect = fig.patch
                rect.set_facecolor('gray')
                rect.set_alpha(0.15)
            else:
                ax = fig.add_subplot(111, aspect='equal', **kwargs)

            # draw room
            polygons = [Polygon(self.corners.T, True)]
            p = PatchCollection(polygons, cmap=matplotlib.cm.jet,
                    facecolor=np.array([1, 1, 1]), edgecolor=np.array([0, 0, 0]))
            ax.add_collection(p)

            # draw the microphones
            if (self.mic_array is not None):
                for mic in self.mic_array.R.T:
                    ax.scatter(mic[0], mic[1],
                            marker='x', linewidth=0.5, s=mic_marker_size, c='k')

                # draw the beam pattern of the beamformer if requested (and available)
                if freq is not None \
                        and isinstance(self.mic_array, bf.Beamformer) \
                        and (self.mic_array.weights is not None or self.mic_array.filters is not None):

                    freq = np.array(freq)
                    if freq.ndim is 0:
                        freq = np.array([freq])

                    # define a new set of colors for the beam patterns
                    newmap = plt.get_cmap('autumn')
                    desat = 0.7
                    try:
                        # this is for matplotlib >= 2.0.0
                        ax.set_prop_cycle(color=[newmap(k) for k in desat*np.linspace(0, 1, len(freq))])
                    except:
                        # keep this for backward compatibility
                        ax.set_color_cycle([newmap(k) for k in desat*np.linspace(0, 1, len(freq))])

                    phis = np.arange(360) * 2 * np.pi / 360.
                    newfreq = np.zeros(freq.shape)
                    H = np.zeros((len(freq), len(phis)), dtype=complex)
                    for i, f in enumerate(freq):
                        newfreq[i], H[i] = self.mic_array.response(phis, f)

                    # normalize max amplitude to one
                    H = np.abs(H)**2/np.abs(H).max()**2

                    # a normalization factor according to room size
                    norm = np.linalg.norm((self.corners - self.mic_array.center), axis=0).max()

                    # plot all the beam patterns
                    i = 0
                    for f, h in zip(newfreq, H):
                        x = np.cos(phis) * h * norm + self.mic_array.center[0, 0]
                        y = np.sin(phis) * h * norm + self.mic_array.center[1, 0]
                        l = ax.plot(x, y, '-', linewidth=0.5)

            # define some markers for different sources and colormap for damping
            markers = ['o', 's', 'v', '.']
            cmap = plt.get_cmap('YlGnBu')
            # draw the scatter of images
            for i, source in enumerate(self.sources):
                # draw source
                ax.scatter(
                    source.position[0],
                    source.position[1],
                    c=cmap(1.),
                    s=20,
                    marker=markers[i %len(markers)],
                    edgecolor=cmap(1.))

                # draw images
                if (img_order is None):
                    img_order = self.max_order

                I = source.orders <= img_order

                val = (np.log2(source.damping[I]) + 10.) / 10.
                # plot the images
                ax.scatter(source.images[0, I],
                    source.images[1, I],
                    c=cmap(val),
                    s=20,
                    marker=markers[i % len(markers)],
                    edgecolor=cmap(val))

            return fig, ax
            
        if(self.dim==3):

            import mpl_toolkits.mplot3d as a3
            import matplotlib.colors as colors
            import matplotlib.pyplot as plt
            import scipy as sp

            fig = plt.figure(figsize=figsize)
            ax = a3.Axes3D(fig)

            # plot the walls
            for w in self.walls:
                tri = a3.art3d.Poly3DCollection([w.corners.T], alpha=0.5)
                tri.set_color(colors.rgb2hex(sp.rand(3)))
                tri.set_edgecolor('k')
                ax.add_collection3d(tri)

            # define some markers for different sources and colormap for damping
            markers = ['o', 's', 'v', '.']
            cmap = plt.get_cmap('YlGnBu')
            # draw the scatter of images
            for i, source in enumerate(self.sources):
                # draw source
                ax.scatter(
                    source.position[0],
                    source.position[1],
                    source.position[2],
                    c=cmap(1.),
                    s=20,
                    marker=markers[i %len(markers)],
                    edgecolor=cmap(1.))

                # draw images
                if (img_order is None):
                    img_order = self.max_order

                I = source.orders <= img_order

                val = (np.log2(source.damping[I]) + 10.) / 10.
                # plot the images
                ax.scatter(source.images[0, I],
                    source.images[1, I],
                    source.images[2, I],
                    c=cmap(val),
                    s=20,
                    marker=markers[i % len(markers)],
                    edgecolor=cmap(val))


            # draw the microphones
            if (self.mic_array is not None):
                for mic in self.mic_array.R.T:
                    ax.scatter(mic[0], mic[1], mic[2],
                            marker='x', linewidth=0.5, s=mic_marker_size, c='k')


            return fig, ax

    def plot_rir(self, FD=False):

        if self.rir is None:
            self.compute_rir()

        import matplotlib.pyplot as plt
        from . import utilities as u

        M = self.mic_array.M
        S = len(self.sources)
        for r in range(M):
            for s in range(S):
                h = self.rir[r][s]
                plt.subplot(M, S, r*S + s + 1)
                if not FD:
                    plt.plot(np.arange(len(h)) / float(self.fs), h)
                else:
                    u.real_spectrum(h)
                plt.title('RIR: mic'+str(r)+' source'+str(s))
                if r == M-1:
                    if not FD:
                        plt.xlabel('Time [s]')
                    else:
                        plt.xlabel('Normalized frequency')


    def add_microphone_array(self, micArray):
        self.mic_array = micArray

    def add_source(self, position, signal=None, delay=0):

        if (not self.is_inside(np.array(position))):
            raise ValueError('The source must be added inside the room.')

        self.sources.append(
                SoundSource(
                    position,
                    signal=signal,
                    delay=delay
                    )
                )

    def first_order_images(self, source_position):

        # projected length onto normal
        ip = np.sum(self.normals * (self.corners - source_position[:, np.newaxis]), axis=0)

        # projected vector from source to wall
        d = ip * self.normals

        # compute images points, positivity is to get only the reflections outside the room
        images = source_position[:, np.newaxis] + 2 * d[:, ip > 0]

        # collect absorption factors of reflecting walls
        damping = (1 - self.absorption[ip > 0])

        # collect the index of the wall corresponding to the new image
        wall_indices = np.arange(len(self.walls))[ip > 0]

        return images, damping, wall_indices


    def image_source_model(self, use_libroom=True):

        self.visibility = []

        for source in self.sources:

            if use_libroom and not libroom_available:
                print("C-extension libroom unavailable. Falling back to pure python")

            # Fall back to pure python if requested or C module unavailable
            if not use_libroom or not libroom_available:
                # Then do it in pure python

                # First, we will generate all the image sources

                # generate first order images
                i, d, w = self.first_order_images(np.array(source.position))
                images = [i]
                damping = [d]
                generators = [-np.ones(i.shape[1])]
                wall_indices = [w]

                # generate all higher order images up to max_order
                o = 1
                while o < self.max_order:
                    # generate all images of images of previous order
                    img = np.zeros((self.dim, 0))
                    dmp = np.array([])
                    gen = np.array([])
                    wal = np.array([])
                    for ind, si, sd in zip(range(images[o-1].shape[1]), images[o - 1].T, damping[o - 1]):
                        i, d, w = self.first_order_images(si)
                        img = np.concatenate((img, i), axis=1)
                        dmp = np.concatenate((dmp, d * sd))
                        gen = np.concatenate((gen, ind*np.ones(i.shape[1])))
                        wal = np.concatenate((wal, w))

                    # sort
                    ordering = np.lexsort(img)
                    img = img[:, ordering]
                    dmp = dmp[ordering]
                    gen = gen[ordering]
                    wal = wal[ordering]

                    if isinstance(self, ShoeBox):
                        '''
                        For shoebox rooms, we can remove duplicate
                        image sources from different wall orderings
                        '''
                        diff = np.diff(img, axis=1)
                        ui = np.ones(img.shape[1], 'bool')
                        ui[1:] = (diff != 0).any(axis=0)

                        # add to array of images
                        images.append(img[:, ui])
                        damping.append(dmp[ui])
                        generators.append(gen[ui])
                        wall_indices.append(wal[ui])

                    else:
                        '''
                        But in general, we have to keep everything
                        '''
                        # add to array of images
                        images.append(img)
                        damping.append(dmp)
                        generators.append(gen)
                        wall_indices.append(wal)

                    # next order
                    o += 1
                    
                o_len = np.array([x.shape[0] for x in generators])
                # correct the pointers for linear structure
                for o in np.arange(2, len(generators)):
                    generators[o] += np.sum(o_len[0:o-1])
                    
                # linearize the arrays
                images_lin = np.concatenate(images, axis=1)
                damping_lin = np.concatenate(damping)
                generators_lin = np.concatenate(generators)
                walls_lin = np.concatenate(wall_indices)
                
                # store the corresponding orders in another array
                ordlist = []
                for o in range(len(generators)):
                    ordlist.append((o+1)*np.ones(o_len[o]))
                orders_lin = np.concatenate(ordlist)

                # add the direct source to the arrays
                source.images = np.concatenate((np.array([source.position]).T, images_lin), axis=1)
                source.damping = np.concatenate(([1], damping_lin))
                source.generators = np.concatenate(([-1], generators_lin+1)).astype(np.int)
                source.walls = np.concatenate(([-1], walls_lin)).astype(np.int)
                source.orders = np.array(np.concatenate(([0], orders_lin)), dtype=np.int)

                # Then we will check the visibilty of the sources
                # visibility is a list with first index for sources, and second for mics
                self.visibility.append([])
                for mic in self.mic_array.R.T:
                    if isinstance(self, ShoeBox):
                        # All sources are visible in shoebox rooms
                        self.visibility[-1].append(np.ones(source.images.shape[1], dtype=bool))
                    else:
                        # In general, we need to check
                        self.visibility[-1].append(
                                self.check_visibility_for_all_images(source, mic, use_libroom=False)
                                )
                self.visibility[-1] = np.array(self.visibility[-1])

                I = np.zeros(self.visibility[-1].shape[1], dtype=bool)
                for mic_vis in self.visibility[-1]:
                    I = np.logical_or(I, mic_vis == 1)

                # Now we can get rid of the superfluous images
                source.images = source.images[:,I]
                source.damping = source.damping[I]
                source.generators = source.generators[I]
                source.walls = source.walls[I]
                source.orders = source.orders[I]

                self.visibility[-1] = self.visibility[-1][:,I]


            else:
                # if libroom is available, use it!

                c_room = self.make_c_room()

                # copy microphone information to struct
                mic = np.asfortranarray(self.mic_array.R, dtype=np.float32)
                c_room.n_microphones = ctypes.c_int(mic.shape[1])
                c_room.microphones = mic.ctypes.data_as(c_float_p)

                src = np.array(source.position, dtype=np.float32)

                if isinstance(self, ShoeBox):

                    # create absorption list in correct order for shoebox algorithm
                    absorption_list_shoebox = np.array([self.absorption_dict[d] for d in self.wall_names], dtype=np.float32)
                    room_dim = np.array(self.shoebox_dim, dtype=np.float32)

                    # Call the dedicated C routine for shoebox room
                    libroom.image_source_shoebox(
                            ctypes.byref(c_room), 
                            src.ctypes.data_as(c_float_p), 
                            room_dim.ctypes.data_as(c_float_p),
                            absorption_list_shoebox.ctypes.data_as(c_float_p),
                            self.max_order
                            )
                else:
                    # Call the general image source generator
                    libroom.image_source_model(ctypes.byref(c_room), src.ctypes.data_as(c_float_p), self.max_order)

                # Recover all the arrays as ndarray from the c struct
                n_sources = c_room.n_sources

                if (n_sources > 0):

                    # numpy wrapper around C arrays
                    images = np.ctypeslib.as_array(c_room.sources, shape=(n_sources, self.dim))
                    orders = np.ctypeslib.as_array(c_room.orders, shape=(n_sources,))
                    gen_walls = np.ctypeslib.as_array(c_room.gen_walls, shape=(n_sources,))
                    attenuations = np.ctypeslib.as_array(c_room.attenuations, shape=(n_sources,))
                    is_visible = np.ctypeslib.as_array(c_room.is_visible, shape=(mic.shape[1], n_sources))

                    # Copy to python managed memory
                    source.images = np.asfortranarray(images.copy().T)
                    source.orders = orders.copy()
                    source.walls = gen_walls.copy()
                    source.damping = attenuations.copy()
                    source.generators = -np.ones(source.walls.shape)

                    self.visibility.append(is_visible.copy())

                    # free the C malloc'ed memory
                    libroom.free_sources(ctypes.byref(c_room))

                    # We need to check that microphones are indeed in the room
                    for m in range(self.mic_array.R.shape[1]):
                        # if not, it's not visible from anywhere!
                        if not self.is_inside(self.mic_array.R[:,m]):
                            self.visibility[-1][m,:] = 0


    def compute_rir(self):
        ''' Compute the room impulse response between every source and microphone '''
        
        self.rir = []

        # Run image source model if this hasn't been done
        if self.visibility is None:
            self.image_source_model()

        for m, mic in enumerate(self.mic_array.R.T):
            h = []
            for s, source in enumerate(self.sources):
                h.append(source.get_rir(mic, self.visibility[s][m], self.fs, self.t0))
            self.rir.append(h)

    def simulate(self, recompute_rir=False):
        ''' Simulates the microphone signal at every microphone in the array '''

        # import convolution routine
        from scipy.signal import fftconvolve

        # Throw an error if we are missing some hardware in the room
        if (len(self.sources) is 0):
            raise ValueError('There are no sound sources in the room.')
        if (self.mic_array is None):
            raise ValueError('There is no microphone in the room.')

        # compute RIR if necessary
        if len(self.rir) == 0 or recompute_rir:
            self.compute_rir()

        # number of mics and sources
        M = self.mic_array.M
        S = len(self.sources)

        # compute the maximum signal length
        from itertools import product
        max_len_rir = np.array([len(self.rir[i][j])
                                for i, j in product(range(M), range(S))]).max()
        f = lambda i: len(
            self.sources[i].signal) + np.floor(self.sources[i].delay * self.fs)
        max_sig_len = np.array([f(i) for i in range(S)]).max()
        L = int(max_len_rir) + int(max_sig_len) - 1
        if L % 2 == 1:
            L += 1

        # the array that will receive all the signals
        signals = np.zeros((M, L))

        # compute the signal at every microphone in the array
        for m in np.arange(M):
            rx = signals[m]
            for s in np.arange(S):
                sig = self.sources[s].signal
                if sig is None:
                    continue
                d = int(np.floor(self.sources[s].delay * self.fs))
                h = self.rir[m][s]
                rx[d:d + len(sig) + len(h) - 1] += fftconvolve(h, sig)

            # add white gaussian noise if necessary
            if self.sigma2_awgn is not None:
                rx += np.random.normal(0., np.sqrt(self.sigma2_awgn), rx.shape)

        # record the signals in the microphones
        self.mic_array.record(signals, self.fs)


    def direct_snr(self, x, source=0):
        ''' Computes the direct Signal-to-Noise Ratio '''

        if source >= len(self.sources):
            raise ValueError('No such source')

        if self.sources[source].signal is None:
            raise ValueError('No signal defined for source ' + str(source))

        if self.sigma2_awgn is None:
            return float('inf')

        x = np.array(x)

        sigma2_s = np.mean(self.sources[0].signal**2)

        d2 = np.sum((x - self.sources[source].position)**2)

        return sigma2_s/self.sigma2_awgn/(16*np.pi**2*d2)


    def get_wall_by_name(self, name):
        '''
        Returns the instance of the wall by giving its name.
        
        :arg name: (string) name of the wall
        
        :returns: (Wall) instance of the wall with this name
        '''
        
        if (name in self.wallsId):
            return self.walls[self.wallsId[name]]
        else:
            raise ValueError('The wall '+name+' cannot be found.')

    def print_wall_sequences(self, source):

        visibilityCheck = np.zeros_like(source.images[0])-1
        
        for imageId in range(len(visibilityCheck)-1, -1, -1):
            print("%2d, %d,%.0f,%.0f --- "%(imageId,source.orders[imageId],source.generators[imageId],source.walls[imageId]), end='')
            p = imageId
            while p >= 0:
                if not np.isnan(source.walls[p]):
                    print(int(source.walls[p]), end='')
                p = source.generators[p]
            print()

    def make_c_room(self):
        ''' Wrapper around the C libroom '''

        # exit if libroom is not available
        if not libroom_available:
            return

        # create the ctypes wall array
        c_walls = (CWALL * len(self.walls))()
        c_walls_array = ctypes.cast(c_walls, c_wall_p)
        for cwall, wall in zip(c_walls_array, self.walls):

            c_corners = wall.corners.ctypes.data_as(c_float_p)

            cwall.dim=wall.dim
            cwall.absorption=wall.absorption
            cwall.normal=(ctypes.c_float * 3)(*wall.normal.tolist())
            cwall.n_corners=wall.corners.shape[1]
            cwall.corners=c_corners

            cwall.origin=(ctypes.c_float * 3)(*wall.plane_point.tolist())

            if wall.dim == 3:
                c_corners_2d = wall.corners_2d.ctypes.data_as(c_float_p)

                cwall.basis=(ctypes.c_float * 6)(*wall.plane_basis.flatten('F').tolist())
                cwall.flat_corners=c_corners_2d

        # create the ctypes Room struct
        c_room = CROOM(
                dim = self.dim,
                n_walls = len(self.walls),
                walls = c_walls_array,
                n_obstructing_walls = self.obstructing_walls.shape[0],
                obstructing_walls = self.obstructing_walls.ctypes.data_as(c_int_p),
                )

        return c_room
        
    def check_visibility_for_all_images(self, source, p, use_libroom=True):
        '''
        Checks visibility from a given point for all images of the given source.
        
        This function tests visibility for all images of the source and returns the results
        in an array.
        
        :arg source: (SoundSource) the sound source object (containing all its images)
        :arg p: (np.array size 2 or 3) coordinates of the point where we check visibility
        
        :returns: (int array) list of results of visibility for each image
            -1 : unchecked (only during execution of the function)
            0 (False) : not visible
            1 (True) : visible
        '''
        
        visibilityCheck = np.zeros_like(source.images[0], dtype=np.int32)-1
        
        if self.is_inside(np.array(p)):
            # Only check for points that are in the room!
            if use_libroom and libroom_available:
                # Call the C routine that checks visibility

                # Create the C struct
                c_room = self.make_c_room()

                # copy source information to struct
                c_room.n_sources = ctypes.c_int(source.images.shape[1])
                c_room.sources = source.images.ctypes.data_as(c_float_p)
                c_room.parents = source.generators.ctypes.data_as(c_int_p)
                c_room.gen_walls = source.walls.ctypes.data_as(c_int_p)
                c_room.orders = source.orders.ctypes.data_as(c_int_p)

                # copy microphone information to struct
                mic = np.array(p, dtype=np.float32)
                c_room.n_microphones = ctypes.c_int(1)
                c_room.microphones = mic.ctypes.data_as(c_float_p)

                # add the array for the visibility information
                c_room.is_visible = visibilityCheck.ctypes.data_as(c_int_p)

                # Run the routine here
                libroom.check_visibility_all(ctypes.byref(c_room))

                return visibilityCheck
            else:
                for imageId in range(len(visibilityCheck)-1, -1, -1):
                    visibilityCheck[imageId] = self.is_visible(source, p, imageId)
        else:
            # If point is outside, nothing is visible
            for imageId in range(len(visibilityCheck)-1, -1, -1):
                visibilityCheck[imageId] = False
            
        return visibilityCheck

            
    def is_visible(self, source, p, imageId = 0):
        '''
        Returns true if the given sound source (with image source id) is visible from point p.
        
        :arg source: (SoundSource) the sound source (containing all its images)
        :arg p: (np.array size 2 or 3) coordinates of the point where we check visibility
        :arg imageId: (int) id of the image within the SoundSource object
        
        :return: (bool)
            False (0) : not visible
            True (1) :  visible
        '''

        p = np.array(p)
        imageId = int(imageId)
        
        # Check if there is an obstruction
        if(self.is_obstructed(source, p, imageId)):
            return False
        
        if (source.orders[imageId] > 0):
        
            # Check if the line of sight intersects the generating wall
            genWallId = int(source.walls[imageId])

            # compute the location of the reflection on the wall
            intersection = self.walls[genWallId].intersection(p, np.array(source.images[:, imageId]))[0]

            # the reflection point needs to be visible from the image source that generates the ray
            if intersection is not None:
                    # Check visibility for the parent image by recursion
                    return self.is_visible(source, intersection, source.generators[imageId])
            else:
                return False
        else:
            return True
      
    def is_obstructed(self, source, p, imageId = 0):
        '''
        Checks if there is a wall obstructing the line of sight going from a source to a point.
        
        :arg source: (SoundSource) the sound source (containing all its images)
        :arg p: (np.array size 2 or 3) coordinates of the point where we check obstruction
        :arg imageId: (int) id of the image within the SoundSource object
        
        :returns: (bool)
            False (0) : not obstructed
            True (1) :  obstructed
        '''
        
        imageId = int(imageId)
        if (np.isnan(source.walls[imageId])):
            genWallId = -1
        else:
            genWallId = int(source.walls[imageId])
        
        # Only 'non-convex' walls can be obstructing
        for wallId in self.obstructing_walls:
        
            # The generating wall can't be obstructive
            if(wallId != genWallId):
            
                # Test if the line segment intersects the current wall
                # We ignore intersections at boundaries of the line of sight
                #intersects, borderOfWall, borderOfSegment = self.walls[wallId].intersects(source.images[:, imageId], p)
                intersectionPoint, borderOfSegment, borderOfWall = self.walls[wallId].intersection(source.images[:, imageId], p)

                if (intersectionPoint is not None and not borderOfSegment):
                    
                    # Only images with order > 0 have a generating wall. 
                    # At this point, there is obstruction for source order 0.
                    if (source.orders[imageId] > 0):

                        imageSide = self.walls[genWallId].side(source.images[:, imageId])
                    
                        # Test if the intersection point and the image are at
                        # opposite sides of the generating wall 
                        # We ignore the obstruction if it is inside the
                        # generating wall (it is what happens in a corner)
                        intersectionPointSide = self.walls[genWallId].side(intersectionPoint)
                        if (intersectionPointSide != imageSide and intersectionPointSide != 0):
                            return True
                    else:
                        return True
                
        return False

    def is_inside(self, p, includeBorders = True):
        '''
        Checks if the given point is inside the room.
        
        :arg p: (np.array dim 2 or 3) point to be tested
        :arg includeBorders: (bool) set true if a point on the wall must be considered inside the room
        
        :returns: (bool) True if the given point is inside the room, False otherwise.
        '''
        
        p = np.array(p)
        if (self.dim != p.shape[0]):
            raise ValueError('Dimension of room and p must match.')
        
        # Compute p0, which is a point outside the room at x coordinate xMin-1
        # (where xMin is the minimum x coordinate among the corners of the walls)
        if (self.dim == 2):
            p0 = np.array([np.amin(np.array([wall.corners[0, :] for wall in self.walls]).flatten())-1, p[1]])
        if (self.dim == 3):
            p0 = np.array([np.amin(np.concatenate([wall.corners[0, :] for wall in self.walls]).flatten())-1, p[1], p[2]])
        
        limitCase = False
        count = 0
        for i in range(len(self.walls)):
            intersects, borderOfWall, borderOfSegment = self.walls[i].intersects(p0, p)
            if borderOfSegment:
                limitCase = True
            if intersects:
                count += 1
        if ((not limitCase and count % 2 == 1) or (limitCase and includeBorders)):
            return True
        else:
            return False


# Room 3D

class ShoeBox(Room):
    '''
    This class extends room for shoebox room in 3D space.
    '''

    def __init__(self, 
            p,
            fs=8000,
            t0=0.,
            absorption=0.,
            max_order=1,
            sigma2_awgn=None,
            sources=None,
            mics=None):

        p = np.array(p, dtype=np.float32)

        if len(p.shape) > 1:
            raise ValueError("p must be a vector of length 2 or 3.")

        self.dim = p.shape[0]

        # if only one point is provided, place the other at origin
        p2 = np.array(p)
        p1 = np.zeros(self.dim)

        # record shoebox dimension in object
        self.shoebox_dim = p2

        # Keep the correctly ordered naming of walls
        # This is the correct order for the shoebox computation later
        # W/E is for axis x, S/N for y-axis, F/C for z-axis
        self.wall_names = ['west', 'east', 'south', 'north']
        if self.dim == 3:
            self.wall_names += ['floor', 'ceiling']

        # copy over the aborption coefficent
        if isinstance(absorption, float):
            self.absorption_dict = dict(zip(self.wall_names, [absorption] * len(self.wall_names)))
            absorption = self.absorption_dict

        self.absorption = []
        if isinstance(absorption, dict):
            self.absorption_dict = absorption
            for d in self.wall_names:
                if d in self.absorption_dict:
                    self.absorption.append(self.absorption_dict[d])
                else:
                    raise KeyError(
                            "Absorbtion needs to have keys 'east', 'west', 'north', 'south', 'ceiling' (3d), 'floor' (3d)"
                            )

            self.absorption = np.array(self.absorption)
        else:
            raise ValueError("Absorption must be either a scalar or a 2x dim dictionnary with entries for 'east', 'west', etc.")


        if self.dim == 2:
            walls = []
            # seems the order of walls is important here, don't change!
            walls.append(Wall(np.array([[p1[0], p2[0]], [p1[1], p1[1]]]), absorption['south'], "south"))
            walls.append(Wall(np.array([[p2[0], p2[0]], [p1[1], p2[1]]]), absorption['east'], "east"))
            walls.append(Wall(np.array([[p2[0], p1[0]], [p2[1], p2[1]]]), absorption['north'], "north"))
            walls.append(Wall(np.array([[p1[0], p1[0]], [p2[1], p1[1]]]), absorption['west'], "west"))

        elif self.dim == 3:
            walls = []
            walls.append(Wall(np.array([[p1[0], p1[0], p1[0], p1[0]], [p2[1], p1[1], p1[1], p2[1]], [p1[2], p1[2], p2[2], p2[2]]]), absorption['west'], "west"))
            walls.append(Wall(np.array([[p2[0], p2[0], p2[0], p2[0]], [p1[1], p2[1], p2[1], p1[1]], [p1[2], p1[2], p2[2], p2[2]]]), absorption['east'], "east"))
            walls.append(Wall(np.array([[p1[0], p2[0], p2[0], p1[0]], [p1[1], p1[1], p1[1], p1[1]], [p1[2], p1[2], p2[2], p2[2]]]), absorption['south'], "south"))
            walls.append(Wall(np.array([[p2[0], p1[0], p1[0], p2[0]], [p2[1], p2[1], p2[1], p2[1]], [p1[2], p1[2], p2[2], p2[2]]]), absorption['north'], "north"))
            walls.append(Wall(np.array([[p2[0], p1[0], p1[0], p2[0]], [p1[1], p1[1], p2[1], p2[1]], [p1[2], p1[2], p1[2], p1[2]]]), absorption['floor'], "floor"))
            walls.append(Wall(np.array([[p2[0], p2[0], p1[0], p1[0]], [p1[1], p2[1], p2[1], p1[1]], [p2[2], p2[2], p2[2], p2[2]]]), absorption['ceiling'], "ceiling"))

        else:
            raise ValueError("Only 2D and 3D rooms are supported.")

        Room.__init__(self, walls, fs, t0, max_order, sigma2_awgn, sources, mics)

    def extrude(self, height):
        ''' Overload the extrude method from 3D rooms '''

        Room.extrude(self, np.array([0., 0., height]))






import ctypes as ct

c_float_p = ct.POINTER(ct.c_float)
c_int_p = ct.POINTER(ct.c_int)

class CWALL(ct.Structure):
    _fields_ = [
            ('dim', ct.c_int),
            ('absorption', ct.c_float),
            ('normal', ct.c_float * 3),
            ('n_corners', ct.c_int),
            ('corners', c_float_p),
            ('origin', ct.c_float * 3),
            ('basis', ct.c_float * 6),
            ('flat_corners', c_float_p),
            ]

c_wall_p = ct.POINTER(CWALL)

class CROOM(ct.Structure):
    _fields_ = [
            ('dim', ct.c_int),
            ('n_walls', ct.c_int),
            ('walls', c_wall_p),
            ('n_sources', ct.c_int),
            ('sources', c_float_p),
            ('parents', c_int_p),
            ('gen_walls', c_int_p),
            ('orders', c_int_p),
            ('attenuations', c_float_p),
            ('n_obstructing_walls', ct.c_int),
            ('obstructing_walls', c_int_p),
            ('n_microphones', ct.c_int),
            ('microphones', c_float_p),
            ('is_visible', c_int_p),
            ]

c_room_p = ct.POINTER(CROOM)




import numpy as np
import os
import platform

from .stft import stft

def median(x, axis=-1, keepdims=False):
    '''
    Computes 95% confidence interval for the median.

    :arg x: (ndarray 1D)

    :returns: A tuple (m, [le, ue]). The confidence interval is [m-le, m+ue].
    '''

    # place the axis on which to compute median in first position
    xsw = np.swapaxes(x, axis, 0)

    # sort the array
    xsw = np.sort(xsw, axis=0)
    n = xsw.shape[0]

    if n % 2 == 1:
        # if n is odd, take central element
        m = xsw[n/2,];
    else:
        # if n is even, average the two central elements
        m = 0.5*(xsw[n/2,] + xsw[n/2+1,]);

    # This table is taken from the Performance Evaluation lecture notes by J-Y Le Boudec
    # available at: http://perfeval.epfl.ch/lectureNotes.htm
    CI = [[1,6],  [1,7],  [1,7],  [2,8],  [2,9],  [2,10], [3,10], [3,11], [3,11],[4,12], \
          [4,12], [5,13], [5,14], [5,15], [6,15], [6,16], [6,16], [7,17], [7,17],[8,18], \
          [8,19], [8,20], [9,20], [9,21], [10,21],[10,22],[10,22],[11,23],[11,23], \
          [12,24],[12,24],[13,25],[13,26],[13,27],[14,27],[14,28],[15,28],[15,29], \
          [16,29],[16,30],[16,30],[17,31],[17,31],[18,32],[18,32],[19,33],[19,34], \
          [19,35],[20,35],[20,36],[21,36],[21,37],[22,37],[22,38],[23,39],[23,39], \
          [24,40],[24,40],[24,40],[25,41],[25,41],[26,42],[26,43],[26,44],[27,44]];
    CI = np.array(CI)

    # Table assumes indexing starting at 1, adjust to indexing from 0
    CI -= 1

    if n < 6:
        # If we have less than 6 samples, we cannot have a confidence interval
        ci = np.zeros((2,) + m.shape)
    elif n <= 70:
        # For 6 <= n <= 70, we use exact values from the table
        j = CI[n-6,0]
        k = CI[n-6,1]
        ci = np.array([xsw[j,]-m ,xsw[k,]-m])
    else:
        # For 70 < n, we use the approximation for large sets
        j = np.floor(0.5*n - 0.98*np.sqrt(n))
        k = np.ceil(0.5*n + 1 + 0.98*np.sqrt(n))
        ci = np.array([xsw[j,]-m,xsw[k,]-m])

    if keepdims:
        m = np.swapaxes(m[np.newaxis,], 0, axis)
        if axis < 0:
            ci = np.swapaxes(ci[:,np.newaxis,], 1, axis)
        else:
            ci = np.swapaxes(ci[:,np.newaxis,], 1, axis+1)

    return m, ci


# Simple mean squared error function
def mse(x1, x2):
    '''
    A short hand to compute the mean-squared error of two signals.

    .. math::

       MSE = \\frac{1}{n}\sum_{i=0}^{n-1} (x_i - y_i)^2


    :arg x1: (ndarray)
    :arg x2: (ndarray)
    :returns: (float) The mean of the squared differences of x1 and x2.
    '''
    return (np.abs(x1-x2)**2).sum()/len(x1)


# Itakura-Saito distance function
def itakura_saito(x1, x2, sigma2_n, stft_L=128, stft_hop=128):

  P1 = np.abs(stft(x1, stft_L, stft_hop))**2
  P2 = np.abs(stft(x2, stft_L, stft_hop))**2

  VAD1 = P1.mean(axis=1) > 2*stft_L**2*sigma2_n
  VAD2 = P2.mean(axis=1) > 2*stft_L**2*sigma2_n
  VAD = np.logical_or(VAD1, VAD2)

  if P1.shape[0] != P2.shape[0] or P1.shape[1] != P2.shape[1]:
    raise ValueError("Error: Itakura-Saito requires both array to have same length")

  R = P1[VAD,:]/P2[VAD,:]

  IS = (R - np.log(R) - 1.).mean(axis=1)

  return np.median(IS)

def snr(ref, deg):

    return np.sum(ref**2)/np.sum((ref-deg)**2)

# Perceptual Evaluation of Speech Quality for multiple files using multiple threads
def pesq(ref_file, deg_files, Fs=8000, swap=False, wb=False, bin='./bin/pesq'):
    '''
    pesq_vals = pesq(ref_file, deg_files, sample_rate=None, bin='./bin/pesq'):
    Computes the perceptual evaluation of speech quality (PESQ) metric of a degraded
    file with respect to a reference file.  Uses the utility obtained from ITU
    P.862 http://www.itu.int/rec/T-REC-P.862-200511-I!Amd2/en

    :arg ref_file:    The filename of the reference file.
    :arg deg_files:   A list of degraded sound files names.
    :arg sample_rate: Sample rates of the sound files [8kHz or 16kHz, default 8kHz].
    :arg swap:        Swap byte orders (whatever that does is not clear to me) [default: False].
    :arg wb:          Use wideband algorithm [default: False].
    :arg bin:         Location of pesq executable [default: ./bin/pesq].

    :returns: (ndarray size 2xN) ndarray containing Raw MOS and MOS LQO in rows 0 and 1, 
        respectively, and has one column per degraded file name in deg_files.
    '''

    if isinstance(deg_files, str):
        deg_files = [deg_files]

    if platform.system() is 'Windows':
        bin = bin + '.exe'

    if not os.path.isfile(ref_file):
        raise ValueError('Some file did not exist')
    for f in deg_files:
        if not os.path.isfile(f):
            raise ValueError('Some file did not exist')

    if Fs not in (8000, 16000):
        raise ValueError('sample rate must be 8000 or 16000')

    args = [ bin, '+%d' % int(Fs) ]

    if swap is True:
        args.append('+swap')

    if wb is True:
        args.append('+wb')

    args.append(ref_file)

    # array to receive all output values
    pesq_vals = np.zeros((2,len(deg_files)))

    # launch pesq for each degraded file in a different process
    import subprocess
    pipes = [ subprocess.Popen(args+[deg], stdout=subprocess.PIPE) for deg in deg_files ]
    states = np.ones(len(pipes), dtype=np.bool)

    # Recover output as the processes finish
    while states.any():

        for i,p in enumerate(pipes):
            if states[i] == True and p.poll() is not None:
                states[i] = False
                out = p.stdout.readlines()
                last_line = out[-1][:-2]

                if wb is True:
                    if not last_line.startswith('P.862.2 Prediction'):
                        raise ValueError(last_line)
                    pesq_vals[:,i] =  np.array([0, float(last_line.split()[-1])])
                else:
                    if not last_line.startswith('P.862 Prediction'):
                        raise ValueError(last_line)
                    pesq_vals[:,i] = np.array(map(float, last_line.split()[-2:]))

    return pesq_vals
